import importlib.util
import logging
import os
from collections.abc import Iterable, Iterator
from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Union

import click
from jinja2 import Environment, FileSystemLoader, Template
from linkml_runtime.dumpers import json_dumper, yaml_dumper
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    Definition,
    DefinitionName,
    Element,
    EnumDefinition,
    SlotDefinition,
    SlotDefinitionName,
    SubsetDefinition,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.generators.erdiagramgen import ERDiagramGenerator
from linkml.generators.plantumlgen import PlantumlGenerator
from linkml.utils.generator import Generator, shared_arguments
from linkml.workspaces.example_runner import ExampleRunner


class MarkdownDialect(Enum):
    python = "python"  # https://python-markdown.github.io/ -- used by mkdocs
    myst = "myst"  # https://myst-parser.readthedocs.io/en/latest/ -- used by sphinx


class DiagramType(Enum):
    mermaid_class_diagram = "mermaid_class_diagram"
    plantuml_class_diagram = "plantuml_class_diagram"
    er_diagram = "er_diagram"


# In future this may become a Union statement, but for now we only have dialects for markdown
DIALECT = MarkdownDialect

MAX_CHARS_IN_TABLE = 80
MAX_RANK = 1000

SCHEMA_SUBFOLDER = "schemas"
CLASS_SUBFOLDER = "classes"
SLOT_SUBFOLDER = "slots"
ENUM_SUBFOLDER = "enums"
TYPE_SUBFOLDER = "types"
SUBSET_SUBFOLDER = "subsets"


def enshorten(input):
    """
    Custom filters to truncate any long text intended to go in a table
    and to remove anything after a newline"""
    if input is None:
        return ""
    if "\n" in input:
        toks = input.split("\n")
        input = toks[0]
    if "." in input:
        toks = input.split(".")
        input = toks[0]
    if len(input) > MAX_CHARS_IN_TABLE - 3:
        input = input[0 : MAX_CHARS_IN_TABLE - 3] + "..."
    return input


def text_to_web(input) -> str:
    """Custom filter to convert multi-line text strings into a suitable format for web/markdown output."""
    if input is None:
        return ""
    return "<br>".join(input.strip().split("\n"))


def _ensure_ranked(elements: Iterable[Element]):
    for x in elements:
        if x.rank is None:
            x.rank = MAX_RANK


@dataclass
class DocGenerator(Generator):
    """
    Generates documentation from a schema

    Note: this is a replacement for MarkdownGenerator

    Documents can be generated using either provided Jinja2 templates, or by
    providing your own

    Currently the provided templates are for markdown but this framework allows
    direct generation to rst, html, etc

    This works via jinja2 templates (found in docgen/ folder). By default, only
    markdown templates are provided. You can either override these, or you can
    create entirely different templates e.g. for html, latex, etc

    The template folder is expected to have files:

        - class.FMT.jinja2
        - enum.FMT.jinja2
        - type.FMT.jinja2
        - slot.FMT.jinja2
        - schema.FMT.jinja2
        - subset.FMT.jinja2
        - index.FMT.jinja2

    Most of these accept a jinja2 variable `element`, except index, schema,
    which accept `schema`. See docgen for examples This will generate a single
    document for every

    - class, enum, type, slot
    - subset
    - imported schema

    It will also create an index file
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["markdown", "rst", "html", "latex"]
    uses_schemaloader = False
    requires_metamodel = False

    # ObjectVars
    dialect: Optional[Union[DIALECT, str]] = None
    """markdown dialect (e.g MyST, Python)"""
    sort_by: str = "name"
    visit_all_class_slots = False
    template_mappings: dict[str, str] = None
    directory: str = None
    """directory in which to write documents"""

    index_name: str = "index"
    """name of the index document"""

    template_directory: str = None
    """directory for custom templates"""

    diagram_type: Optional[Union[DiagramType, str]] = None
    """style of diagram (ER, UML)"""

    include_top_level_diagram: bool = False
    """Whether the index page should include a schema diagram"""

    subfolder_type_separation: bool = False
    """Whether each type (class, slot, etc.) should be put in separate subfolder for navigation purposes"""

    truncate_descriptions: bool = True
    """Whether to truncate long (multi-line) descriptions down to a single line."""

    example_directory: Optional[str] = None
    example_runner: ExampleRunner = field(default_factory=lambda: ExampleRunner())

    genmeta: bool = False
    gen_classvars: bool = True
    gen_slots: bool = True
    no_types_dir: bool = False
    use_slot_uris: bool = False
    use_class_uris: bool = False
    hierarchical_class_view: bool = False
    render_imports: bool = False

    preserve_names: bool = False
    """If true, preserve LinkML element names in docs instead of camelcase/underscore."""

    def __post_init__(self):
        dialect = self.dialect
        if dialect is not None:
            # TODO: simplify this
            if isinstance(dialect, str):
                if dialect == MarkdownDialect.myst.value:
                    dialect = MarkdownDialect.myst
                elif dialect == MarkdownDialect.python.value:
                    dialect = MarkdownDialect.python
                else:
                    raise NotImplementedError(f"{dialect} not supported")
            self.dialect = dialect
        if isinstance(self.diagram_type, str):
            self.diagram_type = DiagramType[self.diagram_type]
        if self.example_directory:
            self.example_runner = ExampleRunner(input_directory=Path(self.example_directory))
        super().__post_init__()
        self.logger = logging.getLogger(__name__)
        self.schemaview = SchemaView(self.schema, merge_imports=self.mergeimports)

        # Override schemaview's get_mappings to use preserved URIs when needed
        if self.preserve_names:
            original_get_mappings = self.schemaview.get_mappings

            def get_mappings_with_preserved_uris(element_name, **kwargs):
                mappings = original_get_mappings(element_name, **kwargs)
                if mappings and element_name:
                    try:
                        element = self.schemaview.get_element(element_name)
                        if element:
                            preserved_uri = self.uri(element, expand=True)
                            # Fix self and native mappings to use preserved URI
                            for key in ("self", "native"):
                                if key in mappings:
                                    mappings[key] = [preserved_uri]
                    except Exception:
                        pass
                return mappings

            self.schemaview.get_mappings = get_mappings_with_preserved_uris

    def serialize(self, directory: str = None) -> None:
        """
        Serialize a schema as a collection of documents

        :param directory: relative or absolute path to directory in which documents are to be written
        :return:
        """
        sv = self.schemaview
        if directory is None:
            directory = self.directory
        if directory is None:
            raise ValueError("Directory must be provided")
        template_vars = {
            "sort_by": self.sort_by,
            "diagram_type": self.diagram_type.value if self.diagram_type else None,
            "include_top_level_diagram": self.include_top_level_diagram,
        }
        self.logger.debug("Processing Index")
        template = self._get_template("index")
        out_str = template.render(gen=self, schema=sv.schema, schemaview=sv, **template_vars)
        self._write(out_str, directory, self.index_name)
        if self._is_single_file_format(self.format):
            self.logger.info(f"{self.format} is a single-page format, skipping non-index elements")
            return
        self.logger.debug("Processing Schemas...")
        template = self._get_template("schema")
        for schema_name in sv.imports_closure():
            self.logger.debug(f"  Generating doc for {schema_name}")
            imported_schema = sv.schema_map.get(schema_name)
            out_str = template.render(gen=self, schema=imported_schema, schemaview=sv, **template_vars)
            self._write(
                out_str,
                f"{directory}/{SCHEMA_SUBFOLDER}" if self.subfolder_type_separation else directory,
                imported_schema.name,
            )
        self.logger.debug("Processing Classes...")
        template = self._get_template("class")
        for cn, c in sv.all_classes().items():
            if self._is_external(c):
                continue
            n = self.name(c)
            self.logger.debug(f"  Generating doc for {n}")
            out_str = template.render(gen=self, element=c, schemaview=sv, **template_vars)
            self._write(out_str, f"{directory}/{CLASS_SUBFOLDER}" if self.subfolder_type_separation else directory, n)
        self.logger.debug("Processing Slots...")
        template = self._get_template("slot")
        for sn, s in sv.all_slots().items():
            if self._is_external(s):
                continue
            n = self.name(s)
            self.logger.debug(f"  Generating doc for {n}")
            s = sv.induced_slot(sn)
            out_str = template.render(gen=self, element=s, schemaview=sv, **template_vars)
            self._write(out_str, f"{directory}/{SLOT_SUBFOLDER}" if self.subfolder_type_separation else directory, n)
        self.logger.debug("Processing Enums...")
        template = self._get_template("enum")
        for en, e in sv.all_enums().items():
            if self._is_external(e):
                continue
            n = self.name(e)
            self.logger.debug(f"  Generating doc for {n}")
            out_str = template.render(gen=self, element=e, schemaview=sv, **template_vars)
            self._write(out_str, f"{directory}/{ENUM_SUBFOLDER}" if self.subfolder_type_separation else directory, n)
        self.logger.debug("Processing Types...")
        template = self._get_template("type")
        for tn, t in sv.all_types().items():
            if self._exclude_type(t):
                continue
            n = self.name(t)
            self.logger.debug(f"  Generating doc for {n}")
            t = sv.induced_type(tn)
            out_str = template.render(gen=self, element=t, schemaview=sv, **template_vars)
            self._write(out_str, f"{directory}/{TYPE_SUBFOLDER}" if self.subfolder_type_separation else directory, n)
        self.logger.debug("Processing Subsets...")
        template = self._get_template("subset")
        for _, s in sv.all_subsets().items():
            if self._is_external(c):
                continue
            n = self.name(s)
            self.logger.debug(f"  Generating doc for {n}")
            out_str = template.render(gen=self, element=s, schemaview=sv, **template_vars)
            self._write(out_str, f"{directory}/{SUBSET_SUBFOLDER}" if self.subfolder_type_separation else directory, n)

    def _write(self, out_str: str, directory: str, name: str) -> None:
        """
        Writes a string in desired format (e.g. markdown) to the appropriate file in a directory

        :param out_str: string to be written
        :param directory: location
        :param name: base name - should correspond to element name
        :return:
        """
        path = Path(directory)
        path.mkdir(parents=True, exist_ok=True)
        file_name = f"{name}.{self._file_suffix()}"
        self.logger.debug(f"  Writing file: {file_name}")
        with open(path / file_name, "w", encoding="UTF-8") as stream:
            stream.write(out_str)

    def _file_suffix(self):
        """
        File suffix to be used for both outputs and template files

        Template files are assumed to be of the form TYPE.FILE_SUFFIX.jinja2
        :return:
        """
        if self.format == "markdown":
            return "md"
        elif self.format == "latex":
            return "tex"
        else:
            return self.format

    def _get_template(self, element_type: str) -> Template:
        """
        Create a jinja2 template object for a given schema element type

        The default location for templates is in the linkml/docgen folder,
        but this can be overridden
        :param element_type: e.g. class, enum, index, subset, ...
        :return:
        """
        if self.template_mappings and element_type in self.template_mappings:
            path = self.template_mappings[element_type]
            # TODO: relative paths
            # loader = FileSystemLoader()
            env = Environment()
            self.customize_environment(env)
            return env.get_template(path)
        else:
            base_file_name = f"{element_type}.{self._file_suffix()}.jinja2"
            folder = None
            if self.template_directory:
                p = Path(self.template_directory) / base_file_name
                if p.is_file():
                    folder = self.template_directory
                else:
                    self.logger.info(
                        f"Could not find {base_file_name} in {self.template_directory} - falling back to default"
                    )
            if not folder:
                package_dir = os.path.dirname(importlib.util.find_spec(__name__).origin)
                folder = os.path.join(package_dir, "docgen", "")
            loader = FileSystemLoader(folder)
            env = Environment(loader=loader)
            self.customize_environment(env)
            return env.get_template(base_file_name)

    def schema_title(self) -> str:
        """
        Returns the title of the schema.

        Uses title field if present, otherwise name

        :return:
        """
        s = self.schemaview.schema
        if s.title:
            return s.title
        else:
            return s.name

    def name(self, element: Element) -> str:
        """
        Returns the name of the element in its canonical form

        :param element: SchemaView element definition
        :return: slot name or numeric portion of CURIE prefixed
            slot_uri
        """
        if element is None:
            return ""
        if self.preserve_names:
            return element.name
        elif type(element).class_name == "slot_definition":
            if self.use_slot_uris:
                curie = self.schemaview.get_uri(element)
                if curie:
                    return curie.split(":")[1]
            return underscore(element.name)
        elif type(element).class_name == "class_definition":
            if self.use_class_uris:
                curie = self.schemaview.get_uri(element)
                if curie:
                    return curie.split(":")[1]
            return camelcase(element.name)
        else:
            return camelcase(element.name)

    def uri(self, element: Element, expand=True) -> str:
        """
        Fetches the URI string for the relevant element

        :param element:
        :return:
        """
        if isinstance(element, SubsetDefinition):
            # Subsets have no uri attribute
            return self.name(element)

        # Simple URI composition for preserved names
        if self.preserve_names and not (self.use_class_uris or self.use_slot_uris):
            base = self.schemaview.schema.id
            if base and ("://" in base or base.startswith("http")):
                sep = "" if base.endswith(("/", "#", "_", ":")) else "/"
                return f"{base}{sep}{self.name(element)}"

        if self.subfolder_type_separation:
            return self.schemaview.get_uri(element, expand=expand, use_element_type=True)

        return self.schemaview.get_uri(element, expand=expand)

    def uri_link(self, element: Union[Element, str]) -> str:
        """Returns a link string (default: markdown links) for a schema element

        :param element: uri string or linkml model element
        :return: hyperlinked markdown or web links
        """
        if isinstance(element, str):
            uri = self.schemaview.expand_curie(element)
            return f"[{element}]({uri})"

        uri = self.uri(element)
        curie = self.uri(element, expand=False)
        return f"[{curie}]({uri})"

    def link_mermaid(self, e: Union[Definition, DefinitionName]) -> str:
        """
        Return link to insert in mermaid diagrams for a given element

        :param e: element to be linked
        :return: string with link
        """
        # Reuse markdown link generation to avoid code duplication.
        md_link = self.link(e)
        if not md_link.endswith(")"):
            return md_link
        link = md_link.rsplit("(")[-1][:-1]
        link = link.removesuffix(".md")
        return f"../{link}/"

    def link(self, e: Union[Definition, DefinitionName], index_link: bool = False) -> str:
        """
        Render an element as a hyperlink

        :param e:
        :param index_link: Whether this is a link for the index page or not
        :return:
        """
        if index_link:
            subfolder = ""
        else:
            subfolder = "../"

        if e is None:
            return "NONE"
        if not isinstance(e, Definition):
            e = self.schemaview.get_element(e)
            if e is None:
                return "NONE"
        if self._is_external(e):
            return self.uri_link(e)
        elif isinstance(e, ClassDefinition):
            if self.use_class_uris:
                curie = self.schemaview.get_uri(e)
                if curie is not None:
                    return self._markdown_link(
                        n=curie.split(":")[1],
                        name=e.name,
                        subfolder=subfolder + CLASS_SUBFOLDER if self.subfolder_type_separation else None,
                    )
            return self._markdown_link(
                self.name(e),
                subfolder=subfolder + CLASS_SUBFOLDER if self.subfolder_type_separation else None,
            )
        elif isinstance(e, EnumDefinition):
            return self._markdown_link(
                self.name(e),
                subfolder=subfolder + ENUM_SUBFOLDER if self.subfolder_type_separation else None,
            )
        elif isinstance(e, SlotDefinition):
            if self.use_slot_uris:
                curie = self.schemaview.get_uri(e)
                if curie is not None:
                    return self._markdown_link(
                        n=curie.split(":")[1],
                        name=e.name,
                        subfolder=subfolder + SLOT_SUBFOLDER if self.subfolder_type_separation else None,
                    )
            return self._markdown_link(
                self.name(e),
                subfolder=subfolder + SLOT_SUBFOLDER if self.subfolder_type_separation else None,
            )
        elif isinstance(e, TypeDefinition):
            return self._markdown_link(
                self.name(e),
                subfolder=subfolder + TYPE_SUBFOLDER if self.subfolder_type_separation else None,
            )
        elif isinstance(e, SubsetDefinition):
            return self._markdown_link(
                self.name(e),
                subfolder=subfolder + SUBSET_SUBFOLDER if self.subfolder_type_separation else None,
            )
        else:
            return e.name

    def links(self, e_list: list[DefinitionName]) -> list[str]:
        """Render list of element documentation pages as hyperlinks.

        :param e_list: list of elements
        :return: list of hyperlinked elements
        """
        return list(map(self.link, e_list))

    def _exclude_type(self, t: TypeDefinition) -> bool:
        return self._is_external(t) and not self.schemaview.schema.id.startswith("https://w3id.org/linkml/")

    def _is_external(self, element: Element) -> bool:
        if element is None:
            return False
        if element.from_schema == "https://w3id.org/linkml/types" and not self.genmeta:
            return True
        else:
            return False

    @staticmethod
    def _markdown_link(n: str, name: str = "", subfolder: str = "") -> str:
        if subfolder:
            rel_path = f"{subfolder}/{n}"
        else:
            rel_path = n

        # if explicit name is provided use that for display name
        if name:
            n = name

        return f"[{n}]({rel_path}.md)"

    def inheritance_tree(self, element: Definition, children: bool = True, **kwargs) -> str:
        """
        Show an element in the context of its is-a hierarchy

        Limitations: currently only implemented for markdown (uses nested bullets)

        :param element: slot or class to be shown
        :param children: if true, show direct children
        :param mixins: if true, show mixins alongside each element
        :param kwargs:
        :return:
        """
        s, depth = self._tree(element, focus=element.name, **kwargs)
        if children:
            if isinstance(element, ClassDefinition):
                all_children = self.schemaview.class_children(element.name, mixins=False)
            else:
                all_children = self.schemaview.slot_children(element.name, mixins=False)
            for c in all_children:
                s += self._tree_info(self.schemaview.get_element(c), depth + 1, **kwargs)
        return s

    def _tree(
        self,
        element: Definition,
        mixins=True,
        descriptions=False,
        focus: DefinitionName = None,
    ) -> tuple[str, int]:
        sv = self.schemaview
        if element.is_a:
            pre, depth = self._tree(
                sv.get_element(element.is_a),
                mixins=mixins,
                descriptions=descriptions,
                focus=focus,
            )
            depth += 1
        else:
            pre, depth = "", 0
        s = pre
        s += self._tree_info(element, depth, mixins=mixins, descriptions=descriptions, focus=focus)
        return s, depth

    def _tree_info(
        self,
        element: Definition,
        depth: int,
        mixins=True,
        descriptions=False,
        focus: DefinitionName = None,
    ) -> str:
        indent = " " * depth * 4

        if self.use_slot_uris or self.use_class_uris:
            name = self.schemaview.get_element(element).name
        else:
            name = self.name(element)

        if element.name == focus:
            lname = f"**{name}**"
        else:
            lname = self.link(element)
        s = f"{indent}* {lname}"
        if mixins and element.mixins:
            s += " ["
            if element.mixins:
                for m in element.mixins:
                    s += f" {self.link(m)}"
            s += "]"
        s += "\n"
        return s

    @staticmethod
    def bullet(e: Element, meta_slot: SlotDefinitionName, backquote=False) -> str:
        """
        Render tag-value for an element as a bullet

        Limitations: currently hardcoded to be markdown
        :param e: element that holds the property, e.g. Person
        :param meta_slot: metamodel property to be shown, e.g. comments
        :param backquote: if true, render as backquote
        :return: formatted string
        """
        v = getattr(e, meta_slot, None)
        if v:
            if backquote:
                v = v.replace("`", "\\`")
                v = f"`{v}`"
            return f"* [{meta_slot}](https://w3id.org/linkml/{meta_slot}): {v}\n"
        else:
            return ""

    @staticmethod
    def number_value_range(e: Union[SlotDefinition, TypeDefinition]) -> str:
        """
        Render the minimum and maximum values for a slot or type as a range, e.g 5-100

        :param e:
        :return:
        """
        r = None
        if isinstance(e, TypeDefinition):
            # TODO: new version
            return None
        if e.minimum_value is not None:
            if e.maximum_value is not None:
                r = f"{e.minimum_value} to {e.maximum_value}"
            else:
                r = f">= {e.minimum_value}"
        else:
            if e.maximum_value is not None:
                r = f"<= {e.maximum_value}"
        return r

    @staticmethod
    def cardinality(slot: SlotDefinition) -> str:
        """
        Render combination of required, multivalued, recommended, and exact_cardinality as a range,
        according to Mermaid conventions. Considers 'required' and 'multivalued' to set defaults
        for 'minimum_cardinality' and 'maximum_cardinality'.

        Reference: https://mermaid.js.org/syntax/classDiagram.html#cardinality-multiplicity-on-relations

        The different cardinality options are:
        - 1 Only 1
        - 0..1 Zero or One
        - 1..* One or more
        - * Many
        - n n (where n>1)
        - 0..n zero to n (where n>1)
        - 1..n one to n (where n>1)
        :param slot: SlotDefinition
        :return: cardinality string as used in Mermaid diagrams
        """
        if slot.exact_cardinality is not None:
            cardinality = str(slot.exact_cardinality)  # handles 'n' case
        else:
            if slot.required or slot.identifier:
                min_card = "1"
            else:
                min_card = str(slot.minimum_cardinality) if slot.minimum_cardinality is not None else "0"

            if slot.multivalued:
                max_card = "*"
            else:
                max_card = str(slot.maximum_cardinality) if slot.maximum_cardinality is not None else "1"

            if min_card == "0":
                if max_card == "1":
                    cardinality = "0..1"  # handles '0..1' case
                elif max_card == "*":
                    cardinality = "*"  # handles '*' case
                else:
                    cardinality = f"0..{max_card}"  # handles '0..n' case
            elif min_card == "1":
                if max_card == "1":
                    cardinality = "1"  # handles '1' case
                elif max_card == "*":
                    cardinality = "1..*"  # handles '1..*' case
                else:
                    cardinality = f"1..{max_card}"  # handles '1..n' case
            else:
                if max_card == "*":
                    cardinality = f"{min_card}..*"  # handles 'n..*' case
                else:
                    cardinality = f"{min_card}..{max_card}"  # handles 'n..m' case

        if slot.recommended:
            cardinality += " _recommended_"

        return cardinality

    def mermaid_directive(self) -> str:
        """
        Writes a mermaid directive. See <https://mermaid-js.github.io/mermaid/#/>_

        This comes after the triple-backtick.

        Note that the directive varies depending on whether the dialect is
        the default python markdown (used by mkdocs) or MyST (used if you
        have a sphinx site)
        """
        if self.dialect is not None and self.dialect == MarkdownDialect.myst:
            return "{mermaid}"
        else:
            return "mermaid"

    def mermaid_diagram(self, class_names: list[Union[str, ClassDefinitionName]] = None) -> str:
        """
        Render a mermaid diagram for a set of classes

        :param class_names:
        :return:
        """
        if self.diagram_type.value == DiagramType.er_diagram.value:
            erdgen = ERDiagramGenerator(self.schemaview.schema, format="mermaid", preserve_names=self.preserve_names)
            if class_names:
                return erdgen.serialize_classes(class_names, follow_references=True, max_hops=2)
            else:
                return erdgen.serialize()
        elif self.diagram_type.value == DiagramType.mermaid_class_diagram.value:
            self.logger.info("This is currently handled in the jinja templates")
        elif self.diagram_type.value == DiagramType.plantuml_class_diagram.value:
            plantumlgen = PlantumlGenerator(self.schema, preserve_names=self.preserve_names)
            plantuml_diagram = plantumlgen.serialize(classes=class_names)
            self.logger.debug(f"Created PlantUML diagram for class: {class_names}")
            return plantuml_diagram
        else:
            raise NotImplementedError(f"Diagram type {self.diagram_type} not implemented")

    @staticmethod
    def latex(text: Optional[str]) -> str:
        """
        Makes text safe for latex

        NOTE: may be incomplete!

        :param text:
        :return:
        """
        if text is None:
            text = ""
        return text.replace("_", "\\_")

    def yaml(self, element: Element, inferred=False) -> str:
        """
        Render element as YAML

        :param element:
        :param inferred: (classes only) show all induced slots as attributes
        :return: yaml string
        """
        if not inferred:
            return yaml_dumper.dumps(element)
        else:
            if not isinstance(element, ClassDefinition):
                raise ValueError(f"Inferred only applicable for classes, not {element.name} {type(element)}")
            # TODO: move this code to schemaview
            c = deepcopy(element)
            attrs = self.schemaview.class_induced_slots(c.name)
            for a in attrs:
                c.attributes[a.name] = a
            c.slots = []
            return yaml_dumper.dumps(c)

    def class_induced_slots(self, class_name: ClassDefinitionName) -> Iterator[SlotDefinition]:
        """
        Yields all induced slots for a class

        Ensures rank is non-null

        :param class_name:
        :return: iterator
        """
        elts = self.schemaview.class_induced_slots(class_name)
        _ensure_ranked(elts)
        yield from elts

    def all_class_objects(self) -> Iterator[ClassDefinition]:
        """
        all class objects in schema

        Ensures rank is non-null
        :return: iterator
        """
        elts = self.schemaview.all_classes(imports=self.render_imports).values()
        _ensure_ranked(elts)
        yield from elts

    def all_slot_objects(self) -> Iterator[SlotDefinition]:
        """
        all slot objects in schema

        Ensures rank is non-null
        :return: iterator
        """
        elts = self.schemaview.all_slots(imports=self.render_imports).values()
        _ensure_ranked(elts)
        yield from elts

    def all_type_objects(self) -> Iterator[TypeDefinition]:
        """
        all type objects in schema

        Ensures rank is non-null
        :return: iterator
        """
        elts = self.schemaview.all_types(imports=self.render_imports).values()
        _ensure_ranked(elts)
        yield from elts

    def all_type_object_names(self) -> list[TypeDefinitionName]:
        return [t.name for t in list(self.all_type_objects())]

    def all_enum_objects(self) -> Iterator[EnumDefinition]:
        """
        all enum objects in schema

        Ensures rank is non-null
        :return: iterator
        """
        elts = self.schemaview.all_enums(imports=self.render_imports).values()
        _ensure_ranked(elts)
        yield from elts

    def all_subset_objects(self) -> Iterator[SubsetDefinition]:
        """
        all enum objects in schema

        Ensures rank is non-null
        :return: iterator
        """
        elts = self.schemaview.all_subsets(imports=self.render_imports).values()
        _ensure_ranked(elts)
        yield from elts

    def class_hierarchy_as_tuples(self) -> Iterator[tuple[int, ClassDefinitionName]]:
        """
        Generate a hierarchical representation of all classes in the schema

        This is represented as a list of tuples (depth: int, cls: ClassDefinitionName),
        where the order is pre-order depth first traversal

        This can then be used to draw a hierarchy within jinja in a number of ways; e.g

        - markdown (by using indentation of " " * depth on a list)
        - tables (by placing fake indentation e.g using underscores)

        Simply iterate through all tuples, drawing each in the appropriate way

        Note: By default all classes are ordered alphabetically

        :return: tuples (depth: int, cls: ClassDefinitionName)
        """
        sv = self.schemaview
        roots = sv.class_roots(mixins=False, imports=self.render_imports)

        # by default the classes are sorted alphabetically
        roots = sorted(roots, key=str.casefold, reverse=True)

        # the stack holds tuples of depth-class that have still to be processed.
        # we seed this with all root classes (which have depth 0)
        # note the stack is processed from the last element first, ie. LIFO
        stack = [(0, root) for root in roots]
        # use iterative depth first traversal
        while len(stack) > 0:
            depth, class_name = stack.pop()
            yield depth, class_name
            children = sorted(
                sv.class_children(class_name=class_name, mixins=False, imports=self.render_imports),
                key=str.casefold,
                reverse=True,
            )
            for child in children:
                # depth first - place at end of stack (to be processed next)
                stack.append((depth + 1, child))

    @staticmethod
    def _is_single_file_format(format: str):
        if format == "latex":
            return True
        else:
            return False

    def inject_slot_info(self, slot: SlotDefinition) -> SlotDefinition:
        """
        Injects additional information into a slot

        TODO: move this functionality into schemaview
        :param slot:
        :return:
        """
        sv = self.schemaview
        if not slot.range:
            slot.range = sv.schema.default_range
        if not slot.range:
            slot.range = "string"
        return slot

    @staticmethod
    def get_direct_slot_names(cls: ClassDefinition) -> list[SlotDefinitionName]:
        """Fetch list of all own attributes of a class, i.e.,
        all slot names of slots that belong to the domain of a class.

        :param cls: class for which we want to determine the attributes
        :return: list of names of all own attributes of a class
        """
        return cls.slots + list(cls.attributes.keys())

    def get_direct_slots(self, cls: ClassDefinition) -> list[SlotDefinition]:
        """Fetch list of all own attributes of a class, i.e.,
        all slots that belong to the domain of a class.

        :param cls: class for which we want to determine the attributes
        :return: list of all own attributes of a class
        """
        return [
            self.inject_slot_info(self.schemaview.induced_slot(sn, cls.name)) for sn in self.get_direct_slot_names(cls)
        ]

    def get_indirect_slots(self, cls: ClassDefinition) -> list[SlotDefinition]:
        """Fetch list of all inherited attributes of a class, i.e.,
        all slots that belong to the domain of a class.

        :param cls: class for which we want to determine the attributes
        :return: list of all own attributes of a class
        """
        sv = self.schemaview
        direct_slot_names = self.get_direct_slot_names(cls)
        return [
            self.inject_slot_info(slot)
            for slot in sv.class_induced_slots(cls.name)
            if slot.name not in direct_slot_names
        ]

    def get_slot_inherited_from(
        self, class_name: ClassDefinitionName, slot_name: SlotDefinitionName
    ) -> list[ClassDefinitionName]:
        """Get the name of the class that a given slot is inherited from.

        :param class_name: name of the class whose slot we are checking
        :param slot_name: name of slot in consideration
        :return: list of classes
        """
        sv = self.schemaview
        induced_slot = sv.induced_slot(slot_name, class_name)
        ancestors = sv.class_ancestors(class_name)
        return list(set(induced_slot.domain_of).intersection(ancestors))

    def get_mixin_inherited_slots(self, cls: ClassDefinition) -> dict[str, list[str]]:
        """Fetch list of all slots acquired through mixing.

        :param cls: class for which we want to determine the mixed in slots
        :return: list of all mixed in slots from each mixin class
        """
        mixed_in_slots = {}
        sv = self.schemaview

        mixins = sv.class_parents(class_name=cls.name, mixins=True, is_a=False)
        for c in mixins:
            mixed_in_slots[c] = sv.class_slots(c)

        return mixed_in_slots

    def example_object_blobs(self, class_name: str) -> list[tuple[str, str]]:
        """Fetch list of all examples of a class.

        :param class_name: class for which we want to determine the examples
        :return: list of all examples of a class
        """
        if not self.example_runner:
            return []
        inputs = self.example_runner.example_source_inputs(class_name)
        objs = []
        for input in inputs:
            stem = Path(input).stem
            with open(input, encoding="utf-8") as f:
                objs.append((stem, f.read()))
        return objs

    def _remove_name_keys(self, obj):
        """Recursively removes 'name' keys from a JSON object representation.

        This is used to clean up the output of ClassRule objects. For example, if we had a rule like:
        rules:
        - title: calibration_standard_if_rt
            preconditions:
                slot_conditions:
                    calibration_target:
                        equals_string: retention_index

        The JSON representation would include redundant "name" keys:
        {
            "slot_conditions": {
                "name": "slot_conditions"
                "calibration_target": {
                    "name": "calibration_target"
                    "equals_string": "retention_index",
                },
            }
        }

        This function removes those redundant "name" keys to make the output cleaner.

        :param obj: The object to clean (dict, list, or primitive value)
        :return: The same object with all "name" keys removed from dicts
        """
        if isinstance(obj, dict):
            return {k: self._remove_name_keys(v) for k, v in obj.items() if k != "name"}
        elif isinstance(obj, list):
            return [self._remove_name_keys(item) for item in obj]
        else:
            return obj

    def classrule_to_dict_view(self, element: ClassDefinition) -> list[dict[str, Any]]:
        """Process all rules (of type ClassRule) asserted on a class.

        This method iterates through all rules asserted on a class and returns a list of
        dictionaries, each containing four pieces of information about a rule:
        _title_, _preconditions_, _postconditions_, and _elseconditions_.
        These values will be read in the jinja template and formatted into a tabular
        view for users to understand the rules applied to the class.

        Note: This method removes redundant "name" keys which are asserted on some
        classes in the Python representation of classes from the metamodel.

        :param element: LinkML class object with `rules` asserted on it
        :return: List of dictionaries with title and "sanitized" conditions for each rule
        """
        if not element.rules:
            return []

        rule_dicts = []

        for rule in element.rules:
            # TODO: expand this list of ClassRule metaslots based on use case
            rule_dict = {
                "title": rule.title or "",
                "preconditions": None,
                "postconditions": None,
                "elseconditions": None,
            }

            for key in ["preconditions", "postconditions", "elseconditions"]:
                condition_obj = getattr(rule, key, None)
                if condition_obj:
                    json_obj = json_dumper.to_dict(condition_obj)
                    sanitized_condition = self._remove_name_keys(json_obj)
                    rule_dict[key] = sanitized_condition

            rule_dicts.append(rule_dict)

        return rule_dicts

    def customize_environment(self, env: Environment):
        if self.truncate_descriptions:
            env.filters["enshorten"] = enshorten
        else:
            env.filters["enshorten"] = text_to_web


@shared_arguments(DocGenerator)
@click.option(
    "--directory",
    "-d",
    required=True,
    help="Path to the directory where you want the Markdown files to be written to",
)
@click.option("--index-name", default="index", show_default=True, help="Name of the index document.")
@click.option("--dialect", help="Dialect or 'flavor' of Markdown used.")
@click.option(
    "--diagram-type",
    type=click.Choice([e.value for e in DiagramType]),
    help="Type of UML diagram to be rendered on class documentation pages.",
)
@click.option(
    "--include-top-level-diagram/--no-include-top-level-diagram",
    default=False,
    show_default=True,
    help="Include ER diagram of the entire schema on index page.",
)
@click.option(
    "--sort-by",
    default="name",
    show_default=True,
    help="Metaslot to use to sort elements by e.g. rank, name, title",
)
@click.option(
    "--genmeta/--no-genmeta",
    default=False,
    show_default=True,
    help="Generating metamodel. Only use this for generating meta.py",
)
@click.option("--template-directory", help="Path to the directory with custom jinja2 templates")
@click.option(
    "--use-slot-uris/--no-use-slot-uris",
    default=False,
    help="Use IDs from slot_uri instead of names",
)
@click.option(
    "--use-class-uris/--no-use-class-uris",
    default=False,
    help="Use IDs from class_uri instead of names",
)
@click.option(
    "--hierarchical-class-view/--no-hierarchical-class-view",
    default=True,
    help="Render class table on index page in a hierarchically indented view",
)
@click.option(
    "--render-imports/--no-render-imports",
    default=False,
    show_default=True,
    help="Render also the documentation of elements from imported schemas",
)
@click.option(
    "--example-directory",
    help="Folder in which example files are found. These are used to make inline examples",
)
@click.option(
    "--include",
    help="""
Include LinkML Schema outside of imports mechanism.  Helpful in including deprecated classes and slots in a separate
YAML, and including it when necessary but not by default (e.g. in documentation or for backwards compatibility)
""",
)
@click.option(
    "--subfolder-type-separation/--no-subfolder-type-separation",
    default=False,
    help="Separate type (class, slot, etc.) outputs in different subfolders for navigation purposes",
)
@click.option(
    "--truncate-descriptions",
    default=True,
    show_default=True,
    help="""
Whether to truncate long (potentially spanning multiple lines) descriptions of classes, slots, etc., in the docs.
Set to true for truncated descriptions, and false to display full descriptions.""",
)
@click.option(
    "--preserve-names/--normalize-names",
    default=False,
    show_default=True,
    help="Preserve original LinkML names in documentation output (e.g., for page titles, links, and file names).",
)
@click.version_option(__version__, "-V", "--version")
@click.command(name="doc")
def cli(
    yamlfile,
    directory,
    index_name,
    dialect,
    template_directory,
    use_slot_uris,
    use_class_uris,
    hierarchical_class_view,
    subfolder_type_separation,
    render_imports,
    truncate_descriptions,
    preserve_names,
    **args,
):
    """Generate documentation folder from a LinkML YAML schema

    Currently a default set of templates for markdown is provided (see the
    folder linkml/generators/docgen/)

    If you specify another format (e.g. html) then you need to provide a
    template_directory argument, with a template for each type of entity inside.

    Examples can optionally be integrated into the documentation; to enable
    this, pass in the --example-directory argument.  The example directory
    should contain one file per example, following the naming convention
    <ClassName>-<ExampleName>.<extension>.

    For example, to include examples on the page for Person, include examples

    Person-001.yaml, Person-002.yaml, etc.

    Currently examples must be in yaml
    """
    gen = DocGenerator(
        yamlfile,
        directory=directory,
        dialect=dialect,
        template_directory=template_directory,
        use_slot_uris=use_slot_uris,
        use_class_uris=use_class_uris,
        hierarchical_class_view=hierarchical_class_view,
        index_name=index_name,
        subfolder_type_separation=subfolder_type_separation,
        render_imports=render_imports,
        truncate_descriptions=truncate_descriptions,
        preserve_names=preserve_names,
        **args,
    )
    print(gen.serialize())


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from typing import Any, Optional, Union

from linkml_runtime.linkml_model import (
    ClassDefinition,
    EnumDefinition,
    PermissibleValue,
    Prefix,
    SchemaDefinition,
    SlotDefinition,
    TypeDefinition,
)
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.schema_as_dict import schema_as_dict


@dataclass
class SchemaBuilder:
    """
    Builder class for SchemaDefinitions.

    Example:

        >>> from linkml.utils.schema_builder import SchemaBuilder
        >>> sb = SchemaBuilder('test-schema')
        >>> sb.add_class('Person', slots=['name', 'age'])
        >>> sb.add_class('Organization', slots=['name', 'employees'])
        >>> sb.add_slot('name',description='Name of the person or organization')
        >>> sb.add_slot('age',description='Age of the person', range='integer')
        >>> schema = sb.schema
        >>> print()

    Most builder methods accepts either a string, an instance of a metamodel element,
    or a dictionary.  If a string is provided, then a new element is created with this
    as the name.

    This follows the standard Builder pattern, so the results of a build operation
    are a builder, allowing chaining. For example:

        >>> sb = SchemaBuilder('test-schema').add_class('Person', slots=['name', 'age'])

    """

    name: Optional[str] = None
    """Initialized name for the schema."""

    id: Optional[str] = None
    """Initialized id for the schema."""

    schema: SchemaDefinition = None
    """generated SchemaDefinition object."""

    def __post_init__(self):
        name = self.name
        if name is None:
            name = "test-schema"
        id = self.id if self.id else f"http://example.org/{name}"
        self.schema = SchemaDefinition(id=id, name=name)

    def add_class(
        self,
        cls: Union[ClassDefinition, dict, str],
        slots: Union[dict, list[Union[str, SlotDefinition]]] = None,
        slot_usage: Union[dict[str, SlotDefinition], dict[str, Any], list[SlotDefinition]] = None,
        replace_if_present=False,
        use_attributes=False,
        **kwargs,
    ) -> "SchemaBuilder":
        """
        Adds a class to the schema.

        :param cls: name, dict object, or ClassDefinition object to add
        :param slots: slot of slot names or slot objects.
        :param slot_usage: slots keyed by slot name
        :param replace_if_present: if True, replace existing class if present
        :param kwargs: additional ClassDefinition properties
        :param use_attributes: if True, add slots as attributes
        :return: builder
        :raises ValueError: if class already exists and replace_if_present=False
        """
        if isinstance(cls, str):
            cls = ClassDefinition(cls, **kwargs)
        if isinstance(cls, dict):
            cls = ClassDefinition(**{**cls, **kwargs})
        if cls.name is self.schema.classes and not replace_if_present:
            raise ValueError(f"Class {cls.name} already exists")
        self.schema.classes[cls.name] = cls
        if use_attributes:
            for s in slots:
                if isinstance(s, SlotDefinition):
                    cls.attributes[s.name] = s
                elif isinstance(s, dict):
                    cls.attributes[s["name"]] = s
                elif isinstance(s, str):
                    cls.attributes[s] = SlotDefinition(s)
                else:
                    raise ValueError("If use_attributes=True then slots must be SlotDefinitions")
        else:
            if slots is not None:
                if isinstance(slots, dict):
                    for k, v in slots.items():
                        cls.slots.append(k)
                        self.add_slot(SlotDefinition(k, **v), replace_if_present=replace_if_present)
                else:
                    for s in slots:
                        cls.slots.append(s.name if isinstance(s, SlotDefinition) else s)
                        if isinstance(s, str) and s in self.schema.slots:
                            # top-level slot already exists
                            continue
                        self.add_slot(s, replace_if_present=replace_if_present)
            if slot_usage:
                if isinstance(slot_usage, dict):
                    for k, v in slot_usage.items():
                        if isinstance(v, dict):
                            v = SlotDefinition(k, **v)
                        cls.slot_usage[k] = v
                elif isinstance(slot_usage, list):
                    for s in slot_usage:
                        cls.slot_usage[s.name] = s
                else:
                    raise ValueError(f"slot_usage {slot_usage} must be a dict or list of SlotDefinitions")
        for k, v in kwargs.items():
            setattr(cls, k, v)
        return self

    def add_slot(
        self,
        slot: Union[SlotDefinition, dict, str],
        class_name: str = None,
        replace_if_present=False,
        **kwargs,
    ) -> "SchemaBuilder":
        """
        Adds the slot to the schema.

        :param slot: name, dict object, or SlotDefinition object to add
        :param class_name: if specified, this will become a valid slot for this class
        :param replace_if_present: if True, replace existing slot if present
        :param kwargs: additional properties
        :return: builder
        :raises ValueError: if slot already exists and replace_if_present=False
        """
        if isinstance(slot, str):
            slot = SlotDefinition(slot, **kwargs)
        elif isinstance(slot, dict):
            slot = SlotDefinition(**{**slot, **kwargs})
        if not replace_if_present and slot.name in self.schema.slots:
            raise ValueError(f"Slot {slot.name} already exists")
        self.schema.slots[slot.name] = slot
        if class_name is not None:
            self.schema.classes[class_name].slots.append(slot.name)
        return self

    def set_slot(self, slot_name: str, **kwargs) -> "SchemaBuilder":
        """
        Set details of the slot

        :param slot_name:
        :param kwargs:
        :return: builder
        """
        slot = self.schema.slots[slot_name]
        for k, v in kwargs.items():
            setattr(slot, k, v)
        return self

    def add_enum(
        self,
        enum_def: Union[EnumDefinition, dict, str],
        permissible_values: list[Union[str, PermissibleValue]] = None,
        replace_if_present=False,
        **kwargs,
    ) -> "SchemaBuilder":
        """
        Adds an enum to the schema

        :param enum_def:
        :param permissible_values:
        :param replace_if_present:
        :param kwargs:
        :return: builder
        :raises ValueError: if enum already exists and replace_if_present=False
        """
        if not isinstance(enum_def, EnumDefinition):
            enum_def = EnumDefinition(enum_def, **kwargs)
        if isinstance(enum_def, dict):
            enum_def = EnumDefinition(**{**enum_def, **kwargs})
        if enum_def.name in self.schema.enums and not replace_if_present:
            raise ValueError(f"Enum {enum_def.name} already exists")
        self.schema.enums[enum_def.name] = enum_def
        if permissible_values is not None:
            for pv in permissible_values:
                if isinstance(pv, str):
                    pv = PermissibleValue(text=pv)
                    enum_def.permissible_values[pv.text] = pv
                else:
                    enum_def.permissible_values[pv.text] = pv
        return self

    def add_prefix(self, prefix: str, url: str, replace_if_present=False) -> "SchemaBuilder":
        """
        Adds a prefix for use with CURIEs

        :param prefix:
        :param url:
        :param replace_if_present:
        :return: builder
        :raises ValueError: if prefix already exists and replace_if_present=False
        """
        obj = Prefix(prefix_prefix=prefix, prefix_reference=url)
        if prefix in self.schema.prefixes and not replace_if_present:
            raise ValueError(f"Prefix {prefix} already exists")
        self.schema.prefixes[obj.prefix_prefix] = obj
        return self

    def add_defaults(self) -> "SchemaBuilder":
        """
        Sets defaults, including:

        - default_range
        - default imports to include linkml:types
        - default prefixes

        :return: builder
        """
        name = underscore(self.schema.name)
        uri = self.schema.id
        self.schema.default_range = "string"
        self.schema.default_prefix = name
        self.schema.imports.append("linkml:types")
        self.add_prefix("linkml", "https://w3id.org/linkml/")
        self.add_prefix(name, f"{uri}/")
        return self

    def add_type(
        self,
        type: Union[TypeDefinition, dict, str],
        typeof: str = None,
        uri: str = None,
        replace_if_present=False,
        **kwargs,
    ) -> "SchemaBuilder":
        """
        Adds the type to the schema

        :param type:
        :param typeof: if specified, the parent type
        :param uri: if specified, the URI or curie of the type
        :param replace_if_present:
        :param kwargs:
        :return: builder
        :raises ValueError: if type already exists and replace_if_present=False
        """
        if isinstance(type, str):
            type = TypeDefinition(type)
        elif isinstance(type, dict):
            type = TypeDefinition(**type)
        if typeof:
            type.typeof = typeof
        if not replace_if_present and type.name in self.schema.types:
            raise ValueError(f"Type {type.name} already exists")
        self.schema.types[type.name] = type
        for k, v in kwargs.items():
            setattr(type, k, v)
        return self

    def as_dict(self) -> dict:
        """
        Returns the schema as a dictionary.

        Compaction is performed to eliminate redundant keys

        :return: dictionary representation
        """
        return schema_as_dict(self.schema)



from .dataframe_class import DataframeClass
from .dataframe_document import DataframeDocument
from .dataframe_field import DataframeField

__all__ = [
    "DataframeDocument",
    "DataframeClass",
    "DataframeField",
]



"""
The ``linkml.validator.plugins`` package contains classes that perform the actual validation work
on data instances. Instances of these classes should be provided when constructing a
:class:`linkml.validator.Validator` instance.
"""

from linkml.validator.plugins.jsonschema_validation_plugin import JsonschemaValidationPlugin
from linkml.validator.plugins.pydantic_validation_plugin import PydanticValidationPlugin
from linkml.validator.plugins.recommended_slots_plugin import RecommendedSlotsPlugin
from linkml.validator.plugins.validation_plugin import ValidationPlugin

__all__ = ["JsonschemaValidationPlugin", "PydanticValidationPlugin", "RecommendedSlotsPlugin", "ValidationPlugin"]



from dataclasses import dataclass
from typing import Union

from linkml_runtime.linkml_model import ClassDefinitionName, SchemaDefinition


@dataclass
class DataValidator:
    """
    Base class for all validators
    """

    schema: Union[str, SchemaDefinition] = None
    """
    LinkML Schema to validate against
    """

    def validate_dict(self, data: dict, target_class: ClassDefinitionName = None, closed: bool = True) -> None:
        raise NotImplementedError



try:
    from importlib import metadata
except ImportError:
    # Python < 3.8
    import importlib_metadata as metadata  # pragma: no cover

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover



"""
This module executes and evaluates code blocks in tutorial markdown files.
Command line usage example: `python -m linkml.utils.execute_tutorial -d /tmp/tutorial/ docs/intro/tutorial01.md`

It requires that the code blocks have the following format:
- A code block containing file contents that should be written to a file for testing should be preceded by a line
whose text is the file name (no spaces) and a colon.
- A code block containing expected output should be preceded by a line whose text is a single word containing the
text "output" (case insensitive) and a colon.
- A code block containing a command that is expected to fail should be preceded by a line that starts with
"<!-- fail" (case insensitive).
- A code block containing commands that should not be executed should be preceded by a line that starts with
"<!-- no_execute" (case insensitive).
- The code block starts with a line containing three backticks. It can then contain either the language name
(e.g., `bash`, `python`, etc.) or `{literalinclude}` followed by a file path. If using `{literalinclude}`,
the next line should contain the language tag (e.g., `:language: python`).
- The code block ends with a line containing three backticks.
"""

import logging
import re
import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path, PurePath

import click

from linkml._version import __version__

logger = logging.getLogger(__name__)

re_decl = re.compile(r"^(\S+):$")
re_start_code_block = re.compile(r"^```(\w+)$")  # does not include literalinclude blocks
re_literalinclude = re.compile(r"^```\{literalinclude\} (.+)$")
re_end_code_block = re.compile(r"^```$")
re_html_comment = re.compile(r"^<!-- (.+) -->")


@dataclass
class Block:
    category: str = None  # yaml, bash, python, ...
    title: str = None
    content: str = None
    output: str = None
    error: str = None
    expected_fail: bool = None
    prior_lines: list[str] = None
    annotations: list[str] = None

    def is_file_block(self) -> bool:
        return self.title and "." in self.title

    def is_stdout(self) -> bool:
        return self.title and "output" in self.title.lower()

    def is_bash(self) -> bool:
        return self.category == "bash"


def execute_blocks(directory: str, blocks: list[Block]) -> list[str]:
    """
    Execute the code blocks embedded in a tutorial

    :param directory:
    :param blocks:
    :return: errors
    """
    Path(directory).mkdir(parents=True, exist_ok=True)
    logger.info(f"Executing in: {directory}")
    last_block = None
    errs = []

    def err(e):
        errs.append(e)
        logger.error(e)

    for block in blocks:
        write_lines(block.prior_lines)
        logger.info(f"# Block: {block.category} {block.title}")
        if block.is_file_block() and not block.is_bash():
            path = PurePath(directory, block.title)
            logger.info(f"Writing to: {path}")
            with open(path, "w", encoding="UTF-8") as stream:
                stream.write(block.content)
        elif block.is_bash():
            if "no_execute" in block.annotations:
                continue
            cmd = block.content.strip().split()
            if ">" in cmd:
                # redirects not support in subprocess.run
                pos = cmd.index(">")
                outpath = cmd[pos + 1 :]
                cmd = cmd[0:pos]
                if len(outpath) > 1:
                    raise Exception(f"Maximum 1 token after > in {block.content}. Got: {outpath}")
                outpath = str(Path(directory, *outpath))
                logger.info(f"OUTPATH = {outpath}")
            else:
                outpath = None
            logger.info(f"Executing: {cmd}")
            r = subprocess.run(cmd, cwd=directory, capture_output=True)
            block.output = r.stdout.decode("utf-8")
            if outpath:
                with open(outpath, "w", encoding="UTF-8") as stream:
                    logger.info(f"WRITING {len(block.output)} CHARS TO = {outpath}")
                    stream.write(block.output)
            block.error = r.stderr.decode("utf-8")
            logger.info(f"OUT [sample] = {block.output[0:30]}")
            if block.expected_fail:
                if r.returncode == 0:
                    err(f"Command unexpectedly succeeded: {cmd}")
                else:
                    logger.info("Failed as expected")
                if block.error:
                    logger.info(f"ERR [sample] = ...{block.error[-200:]}")
            else:
                if block.error:
                    logger.info(f"ERR = {block.error}")
                if r.returncode != 0:
                    err(f"Command failed: {cmd}")
                else:
                    logger.info("Success!")
        elif block.is_stdout():
            if "compare_rdf" in block.annotations:
                logger.warning("SKIPPING RDF COMPARISON. TODO: https://github.com/linkml/linkml/issues/427")
            elif last_block.output:
                if last_block.output.strip() != block.content.strip():
                    err(f"Mismatch: {str(last_block.output)} != {block.content}")
                else:
                    logger.info("Hurray! Contents match!")
            else:
                logger.info("No comparison performed")
        else:
            logger.warning(f"Ignoring block: {block}")
        last_block = block
    return errs


def write_lines(lines: list[str]) -> None:
    for line in lines:
        print(f"+++ {line}")


def parse_file_to_blocks(input) -> list[Block]:
    """
    Parses a markdown tutorial file to code blocks to be executed

    :param input:
    :return:
    """
    with open(input) as stream:
        lines = stream.readlines()
    curr_block = None
    blocks = []
    fn = None
    prior_lines = []
    anns = []
    expected_fail = False
    while lines:
        line = lines[0].rstrip()
        lines = lines[1:]

        if curr_block is None:
            if line.lower().startswith("<!-- fail"):
                expected_fail = True
            m = re_html_comment.match(line)
            if m:
                anns.append(m.group(1).strip().lower())
                continue
            m = re_decl.match(line)
            if m:
                fn = m.group(1)
                print(f"TITLE={fn}")
            else:
                m = re_start_code_block.match(line)
                if m:
                    curr_block = Block(
                        category=m.group(1),
                        title=fn,
                        expected_fail=expected_fail,
                        content="",
                        annotations=anns,
                        prior_lines=prior_lines,
                    )
                    prior_lines = []
                    anns = []
                else:
                    m = re_literalinclude.match(line)
                    if m and fn:
                        # Handle literalinclude directive
                        file_path = m.group(1).strip()
                        # Look ahead and extract language from the language tag
                        language = "yaml"  # Default to yaml if no language tag is found
                        if lines and lines[0].strip().startswith(":language:"):
                            language_line = lines[0].strip()
                            language = language_line.split(":language:")[1].strip()
                            lines = lines[1:]
                        # Skip end of code block
                        if lines and lines[0].strip() == "```":
                            lines = lines[1:]
                        else:
                            logger.warning(
                                f"Expected end of code block after literalinclude, but found: {lines[0].strip()}"
                            )

                        # Convert relative path to absolute path
                        input_dir = Path(input).parent
                        abs_path = input_dir / file_path

                        try:
                            with open(abs_path) as f:
                                file_content = f.read()

                            # Create a block for this file
                            file_block = Block(
                                category=language,  # Use the extracted language
                                title=fn,
                                expected_fail=expected_fail,
                                content=file_content,
                                annotations=anns,
                                prior_lines=prior_lines,
                            )
                            blocks.append(file_block)
                            prior_lines = []
                            anns = []
                        except Exception as e:
                            logger.error(f"Failed to read included file {abs_path}: {e}")
                    else:
                        prior_lines += [line]
        else:
            # in block
            fn = None
            expected_fail = False
            m = re_end_code_block.match(line)
            if m:
                blocks.append(curr_block)
                curr_block = None
            else:
                curr_block.content += f"{line}\n"
    return blocks


@click.command()
@click.option(
    "--directory",
    "-d",
    required=True,
    help="path to directory to execute tutorial example on",
)
@click.argument("inputs", nargs=-1)
@click.version_option(__version__, "-V", "--version")
def cli(inputs, directory):
    """
    Execute a tutorial markdown file (eg. those in the /docs/intro/ directory) and
    save the outputs in the given directory

    Example:

        export PYTHONPATH=`pwd`
        python -m linkml.utils.execute_tutorial -d /tmp/tutorial/ docs/intro/tutorial01.md

    """
    logging.basicConfig(level=logging.INFO)
    errs = []
    for input in inputs:
        logger.info(f"INPUT={input}")
        blocks = parse_file_to_blocks(input)
        print(f"## {len(blocks)} Blocks")
        localdir = Path(input).stem
        subdir = PurePath(directory, localdir)
        input_errs = execute_blocks(str(subdir), blocks)
        if len(input_errs) > 0:
            logger.error(f"TUTORIAL {input} FAILURES: {len(input_errs)}")
        errs += input_errs
    logger.info(f"Errors = {len(errs)}")
    if len(errs) > 0:
        logger.error(f"Encountered {len(errs)} Errors")
        for err in errs:
            logger.error(f"Error: {err}")
        sys.exit(1)


if __name__ == "__main__":
    cli(sys.argv[1:])



from abc import ABC, abstractmethod


class ModelTransform(ABC):
    @abstractmethod
    def explode_unnest_dataframe(self, df, column_name):
        """Abstract method for exploding and unnesting dataframes."""
        pass



import csv
import re
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Optional

from linkml.validator.loaders.loader import Loader


def _parse_numeric(value: str):
    if not isinstance(value, str) or not re.search(r"[0-9]", value):
        return value
    try:
        return int(value)
    except (TypeError, ValueError):
        pass
    try:
        return float(value)
    except (TypeError, ValueError, OverflowError):
        return value


class _DelimitedFileLoader(Loader, ABC):
    """Base class for TSV and CSV loaders"""

    @property
    @abstractmethod
    def delimiter(self):
        pass

    def __init__(self, source, *, skip_empty_rows: bool = False, index_slot_name: Optional[str] = None) -> None:
        super().__init__(source)
        self.skip_empty_rows = skip_empty_rows
        self.index_slot_name = index_slot_name

    def _rows(self) -> Iterator[dict]:
        with open(self.source) as file:
            reader: csv.DictReader = csv.DictReader(file, delimiter=self.delimiter, skipinitialspace=True)
            for row in reader:
                if self.skip_empty_rows and not any(row.values()):
                    continue
                yield {k: _parse_numeric(v) for k, v in row.items() if k is not None and v != ""}

    def iter_instances(self) -> Iterator[dict]:
        if self.index_slot_name is not None:
            yield {self.index_slot_name: list(self._rows())}
        else:
            yield from self._rows()


class CsvLoader(_DelimitedFileLoader):
    """A loader for instances serialized as CSV

    :param skip_empty_rows: If ``True``, skip empty rows instead of yielding empty dicts. Defaults
        to ``False``.
    :param index_slot_name: If provided, ``iter_instances`` will yield one dict where all rows of
        the CSV file are collected into a list with ``index_slot_name`` as the key. If ``None``,
        ``iter_instances`` will yield each row as a dict individually. Defaults to ``None``.
    """

    @property
    def delimiter(self):
        return ","


class TsvLoader(_DelimitedFileLoader):
    """A loader for instances serialized as TSV

    :param skip_empty_rows: If ``True``, skip empty rows instead of yielding empty dicts. Defaults
        to ``False``.
    :param index_slot_name: If provided, ``iter_instances`` will yield one dict where all rows of
        the TSV file are collected into a list with ``index_slot_name`` as the key. If ``None``,
        ``iter_instances`` will yield each row as a dict individually. Defaults to ``None``.
    """

    @property
    def delimiter(self):
        return "\t"



import polars as pl

from .model_transform import ModelTransform


class ListDictModelTransform(ModelTransform):
    """This class assists in converting a LinkML 'list dict' inline column
    into a form that is better for representing in a PolaRS dataframe and
    validating with a Pandera model.
    """

    def __init__(self, polars_schema):
        self.polars_schema = polars_schema
        """A polars schema representing a list dict column"""

    def transform(self, linkml_list_dict):
        """Transforms a list dict nested column.
        This is a pass-through since list dicts are already in the correct format.
        """
        return linkml_list_dict

    @classmethod
    def unnest_list_struct(cls, column_name: str, df):
        """Use this in a custom check. Pass the nested model as pandera_model."""

        # fmt: off
        unnested_column = (
            df
            .select(column_name)
            .filter(pl.col(column_name).list.len() > 0) # see: https://github.com/pola-rs/polars/issues/14381
            .explode(column_name)
            .unnest(column_name)
        )
        # fmt: on

        return unnested_column

    @classmethod
    def prepare_dataframe(cls, data, column_name, nested_cls):
        """Returns just the list dict column transformed to an inlined list form

        note that this method uses collect and iter_rows so is very inefficient
        """
        polars_schema = nested_cls.to_schema()

        list_dict_transformer = cls(polars_schema)

        one_column_df = data.lazyframe.select(pl.col(column_name)).collect()

        list_of_structs = [list_dict_transformer.transform(e) for [e] in one_column_df.iter_rows()]

        return pl.DataFrame(pl.Series(list_of_structs).alias(column_name))

    def explode_unnest_dataframe(self, df, column_name, data=None):
        """Filter, explode and unnest for list dict with struct fallback."""
        try:
            return (
                df.lazy().filter(pl.col(column_name).list.len() > 0).explode(column_name).unnest(column_name).collect()
            )
        except (pl.exceptions.PanicException, Exception):
            if data:
                from .nested_struct_model_transform import NestedStructModelTransform

                nested_transform = NestedStructModelTransform(self.polars_schema)
                return nested_transform.explode_unnest_dataframe(data.lazyframe, column_name)
            raise



from dataclasses import dataclass

from linkml_runtime.linkml_model.meta import (
    ClassDefinitionName,
    ElementName,
    EnumDefinitionName,
    SlotDefinitionName,
    SubsetDefinitionName,
    TypeDefinitionName,
)
from linkml_runtime.utils.metamodelcore import empty_set


@dataclass(repr=False, frozen=True)
class RefType:
    name: str

    def __repr__(self):
        return self.name


ClassType = RefType("Class")
TypeType = RefType("Type")
SlotType = RefType("Slot")
SubsetType = RefType("Subset")
EnumType = RefType("Enum")


@dataclass
class References:
    """
    Summary of references to a given class. The reference class is the key to the dictionary carrying classrefs
    """

    classrefs: set[ClassDefinitionName] = empty_set()  # Refs of type class
    slotrefs: set[SlotDefinitionName] = empty_set()  # Refs of type slot
    typerefs: set[TypeDefinitionName] = empty_set()  # Refs of type type
    subsetrefs: set[SubsetDefinitionName] = empty_set()  # Refs of type subset
    enumrefs: set[EnumDefinitionName] = empty_set()  # Refs of type enum

    def addref(self, fromtype: RefType, fromname: ElementName) -> None:
        if fromtype is ClassType:
            self.classrefs.add(ClassDefinitionName(fromname))
        elif fromtype is TypeType:
            self.typerefs.add(TypeDefinitionName(fromname))
        elif fromtype is SlotType:
            self.slotrefs.add(SlotDefinitionName(fromname))
        elif fromtype is SubsetType:
            self.subsetrefs.add(SubsetDefinitionName(fromname))
        elif fromtype is EnumType:
            self.slotrefs.add(EnumDefinitionName(fromname))
        else:
            raise TypeError(f"Unknown typ: {fromtype}")

    def update(self, other: "References") -> None:
        self.classrefs.update(other.classrefs)
        self.slotrefs.update(other.slotrefs)
        self.typerefs.update(other.typerefs)
        self.subsetrefs.union(other.subsetrefs)
        self.enumrefs.update(other.enumrefs)

    def __bool__(self):
        return (
            bool(self.classrefs)
            or bool(self.slotrefs)
            or bool(self.typerefs)
            or bool(self.subsetrefs)
            or bool(self.enumrefs)
        )



import csv
import io
from typing import IO, Any, Optional

from ..linter import LinterProblem
from .formatter import Formatter


class TsvFormatter(Formatter):
    def __init__(self, file: Optional[IO[Any]] = None) -> None:
        super().__init__(file)
        self.output = io.StringIO()
        self.writer = csv.writer(self.output, delimiter="\t", quoting=csv.QUOTE_MINIMAL)

    def start_report(self):
        self.writer.writerow(["source", "schema name", "rule name", "level", "message"])

    def handle_problem(self, problem: LinterProblem):
        self.writer.writerow(
            [
                problem.schema_source,
                problem.schema_name,
                problem.rule_name,
                str(problem.level),
                problem.message,
            ]
        )

    def end_report(self):
        self.write(self.output.getvalue())



import polars as pl


class SimpleDictLoader:
    def __init__(
        self,
        struct_schema,
        id_col="id",
        other_col="other",
        id_dtype=pl.String,
        other_dtype=pl.Int64,
        nested_tx=lambda x: x,
    ):
        self.struct_schema = struct_schema
        self.id_col: str = id_col
        self.other_col: str = other_col
        self.id_dtype = id_dtype
        self.other_dtype = other_dtype
        self.nested_tx = nested_tx
        self.polars_schema_keys = set(self.struct_schema.keys())
        self.ordered_schema_keys = list(self.struct_schema.keys())

    def tx_core(self, linkml_simple_dict):
        """core simple dict to list of dicts logic"""
        for id_value, range_value in linkml_simple_dict.items():
            base_dict = {k: None for k in self.ordered_schema_keys}
            if isinstance(range_value, dict) and (set(range_value.keys()) <= self.polars_schema_keys):
                base_dict.update(self.nested_tx(range_value))
            else:
                base_dict[self.other_col] = range_value
            base_dict[self.id_col] = id_value
            yield base_dict

    # simple dict handling
    def tx(self, sd):
        """simple dict to list of dicts"""
        return self.tx_core(sd)

    def tx_batch(self, series):
        """Process entire series of simple dicts"""
        result = []
        for simple_dict in series:
            transformed = list(self.tx(simple_dict))
            result.append(transformed)
        return pl.Series(result, dtype=pl.List(pl.Struct(self.struct_schema)))

    def load(self, source_col):
        return pl.col(source_col).map_batches(self.tx_batch, return_dtype=pl.List(pl.Struct(self.struct_schema)))

    def load_df(self, df, source_col):
        return df.with_columns(self.load(source_col))



import importlib
import logging
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import click
from jinja2 import Environment, FileSystemLoader
from linkml_runtime.linkml_model.meta import Element, SlotDefinition
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml.generators.docgen import DocGenerator
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class MermaidClassDiagramGenerator(Generator):
    """This generator creates Mermaid Class diagrams (https://mermaid.js.org/syntax/classDiagram.html)
    for individual classes in a given LinkML schema and outputs them as Markdown files. It uses
    a jinja template to render these class diagrams, which can be customized by the user and pointed
    to using the `--template-file` option. If no template is provided, then the default template
    file provided by LinkML and present at linkml/generators/docgen/class_diagram.md.jinja2 will be used.
    The generator also has the option of specifying a certain set of classes for which you want the
    class diagrams to be generated by using the `--classes` option.
    """

    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["markdown"]
    uses_schemaloader = False
    requires_metamodel = False

    directory: Optional[str] = None  # output directory with generated markdown files
    template_file: Optional[str] = None  # custom/default jinja template for class diagrams
    classes: list[str] = field(default_factory=list)  # optional subset of classes
    preserve_names: bool = False  # preserve original LinkML names in diagram output

    def __post_init__(self):
        super().__post_init__()
        self.logger = logging.getLogger(__name__)
        self.schemaview = SchemaView(self.schema, merge_imports=self.mergeimports)

        # set the output directory
        self.output_directory = Path(self.directory)

        # set the template file
        if not self.template_file:
            package_dir = os.path.dirname(importlib.util.find_spec("linkml").origin)
            # Default location of the template file
            self.template_file = os.path.join(package_dir, "generators", "docgen", "class_diagram.md.jinja2")

    def generate_class_diagrams(self):
        """Generate Mermaid class diagrams for the specified subset of classes
        or all classes if none are specified.
        """
        self.output_directory.mkdir(parents=True, exist_ok=True)

        template_folder = os.path.dirname(self.template_file)
        template_name = os.path.basename(self.template_file)
        loader = FileSystemLoader(template_folder)
        env = Environment(loader=loader)
        temp_doc_gen = DocGenerator(self.schema, mergeimports=self.mergeimports, preserve_names=self.preserve_names)
        temp_doc_gen.customize_environment(env)

        template = env.get_template(template_name)

        all_classes = self.schemaview.all_classes()

        if self.classes:
            class_items = [(cn, all_classes[cn]) for cn in self.classes if cn in all_classes]
        else:
            class_items = list(all_classes.items())

        for cn, class_def in class_items:
            self.logger.info(f"Generating Mermaid diagram for class: {cn}")
            rendered = template.render(gen=self, element=class_def, schemaview=self.schemaview)
            filename = self.name(class_def) if self.preserve_names else cn
            outfile = self.output_directory / f"{filename}.md"
            with open(outfile, "w", encoding="utf-8") as f:
                f.write(rendered)

    def cardinality(self, slot: SlotDefinition) -> str:
        """Reuses the cardinality logic from DocGenerator."""
        return DocGenerator.cardinality(slot)

    def mermaid_directive(self) -> str:
        """Provides the code fence directive (e.g., `mermaid` or `{mermaid}`)."""
        return "mermaid"

    def name(self, element: Element) -> str:
        """Returns the canonical name for an element."""
        if element is None:
            return ""
        if self.preserve_names:
            return element.name
        elif type(element).class_name == "slot_definition":
            return underscore(element.name)
        elif type(element).class_name == "class_definition":
            return camelcase(element.name)
        else:
            return camelcase(element.name)

    def link_mermaid(self, element):
        """Generates a link for the given element."""
        return f"../{self.name(element)}"

    def all_type_object_names(self):
        return list(self.schemaview.all_types().keys())


@shared_arguments(MermaidClassDiagramGenerator)
@click.command()
@click.option(
    "--template-file",
    "-t",
    type=click.Path(exists=True),
    default=None,
    help="Path to Jinja template for class diagrams.",
)
@click.option(
    "--directory",
    "-d",
    type=click.Path(),
    required=True,
    help="Folder in which to write the Markdown files.",
)
@click.option(
    "--classes",
    "-c",
    multiple=True,
    help="One or more classes in the schema for which to generate diagrams. "
    "If omitted, diagrams for all classes are generated.",
)
@click.option(
    "--preserve-names/--normalize-names",
    default=False,
    show_default=True,
    help="Preserve original LinkML names in Mermaid diagram output (e.g., for class names, slot names, file names).",
)
@click.version_option(click.__version__, "-V", "--version")
def cli(yamlfile, template_file, directory, classes, **args):
    logging.basicConfig(level=logging.INFO)
    gen = MermaidClassDiagramGenerator(
        schema=yamlfile,
        template_file=template_file,
        directory=directory,
        classes=list(classes),  # convert tuple to a list
        **args,
    )
    gen.generate_class_diagrams()


if __name__ == "__main__":
    cli()



"""Iterate through all examples in a folder testing them for validity."""

import glob
import json
import logging
import os
import sys
from collections.abc import Mapping
from dataclasses import dataclass, field
from io import StringIO
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, TextIO, Union

import click
import yaml
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import json_dumper, rdflib_dumper, yaml_dumper
from linkml_runtime.linkml_model import ElementName
from linkml_runtime.utils.formatutils import camelcase

from linkml._version import __version__
from linkml.generators.pythongen import PythonGenerator
from linkml.utils.helpers import get_range_associated_slots
from linkml.validator import Validator, _get_default_validator

logger = logging.getLogger(__name__)


@dataclass
class SummaryDocument:
    """
    An object describing the summary for processing a set of examples
    """

    text: StringIO = field(default_factory=lambda: StringIO())

    inputs: list[str] = field(default_factory=list)

    outputs: list[str] = field(default_factory=list)

    def add(self, *lines: str):
        for line in lines:
            self.text.write(line)
            self.text.write("\n")

    def __str__(self) -> str:
        return self.text.getvalue()


@dataclass
class ExampleRunner:
    """
    Processes a collection of positive and negative examples.

    Background: https://github.com/linkml/linkml/issues/501
    """

    input_directory: Optional[Path] = None
    """Directory in which positive instance examples are found."""

    input_formats: Optional[list[str]] = field(default_factory=lambda: ["yaml"])

    counter_example_input_directory: Optional[Path] = None
    """Directory in which negative instance examples are found. These are expected to fail."""

    output_directory: Optional[Path] = None
    """Directory where processed examples are written to."""

    output_formats: Optional[list[str]] = field(default_factory=lambda: ["yaml", "json", "ttl"])

    schemaview: Optional[SchemaView] = None
    """View over schema which all examples adhere to."""

    summary: SummaryDocument = field(default_factory=lambda: SummaryDocument())

    _python_module: Optional[ModuleType] = None
    """Module containing classes that all examples instantiate."""

    prefix_map: Optional[Mapping[str, str]] = None
    """Custom prefix map, for emitting RDF/turtle."""

    _validator: Optional[Validator] = None

    expand_dicts: bool = None
    """If true, then expand all dicts prior to validation."""

    use_type_designators: bool = True

    @property
    def python_module(self) -> ModuleType:
        """
        Load the python module containing the classes that all examples instantiate.

        :return:
        """
        if self._python_module is None:
            # See: https://github.com/linkml/linkml/issues/1219
            src = self.schemaview.schema.source_file
            if not src:
                src = self.schemaview.schema
            pygen = PythonGenerator(src)
            self._python_module = pygen.compile_module()
        return self._python_module

    @property
    def validator(self) -> Validator:
        """
        Get the current validator

        :return:
        """
        if self._validator is None:
            self._validator = _get_default_validator(self.schemaview.schema)
        return self._validator

    def process_examples(self):
        """
        Process all examples in the input directory.

        Filenames should be of the form

         - CLASSNAME-EXAMPLENAME.yaml

         E.g Person-001.yaml

        :return:
        """
        input_dir = self.input_directory
        counter_example_dir = self.counter_example_input_directory
        if input_dir is None:
            input_dir = Path.cwd() / "examples"
        if counter_example_dir is None:
            counter_example_dir = Path.cwd() / "counter_examples"
        for fmt in self.input_formats:
            input_examples = glob.glob(os.path.join(str(input_dir), f"*.{fmt}"))
            input_counter_examples = glob.glob(os.path.join(str(counter_example_dir), f"*.{fmt}"))
            if not input_counter_examples:
                logger.warning(f"No counter examples found in {self.counter_example_input_directory}")
            self.process_examples_from_list(input_examples, fmt, False)
            self.process_examples_from_list(input_counter_examples, fmt, True)

    def example_source_inputs(self, class_name: str = None) -> list[str]:
        """
        Get the list of example source inputs.

        :param class_name:
        :return:
        """
        input_dir = self.input_directory
        if input_dir is None:
            return []
        all_inputs = []
        for fmt in self.input_formats:
            glob_expr = f"*.{fmt}"
            if class_name is not None:
                glob_expr = f"{class_name}-{glob_expr}"
            input_examples = glob.glob(os.path.join(str(input_dir), glob_expr))
            all_inputs.extend(input_examples)
        return all_inputs

    def process_examples_from_list(self, input_examples: list, input_format: str, counter_examples: bool = True):
        sv = self.schemaview
        validator = self.validator
        summary = self.summary
        for input_example in input_examples:
            stem = Path(input_example).stem
            base = Path(self.output_directory) / stem
            if base in summary.inputs:
                raise ValueError(f"Duplicate example: {base}")
            summary.inputs.append(str(stem))
            base.parent.mkdir(exist_ok=True, parents=True)
            parts = stem.split("-")
            tc = parts[0]
            with open(input_example) as file:
                if input_format == "yaml":
                    input_dict = yaml.safe_load(file)
                elif input_format == "json":
                    input_dict = json.load(file)
                else:
                    raise NotImplementedError(f"Cannot handle format: {input_format}")
                summary.add(f"## {stem}", "### Input", "```yaml", f"{yaml.dump(input_dict)}", "```")
                success = True
                try:
                    report = validator.validate(input_dict, tc)
                    if report.results:
                        raise Exception(
                            "\n".join(f"[{result.severity.value}] {result.message}" for result in report.results)
                        )
                    # json validation is incomplete: also try object instantiation
                    self._load_from_dict(input_dict, target_class=tc)
                except Exception as e:
                    success = False
                    if not counter_examples:
                        raise ValueError(f"Example {input_example} failed validation:\n{e}")
                if counter_examples:
                    if success:
                        raise ValueError(f"Counter example {input_example} succeeded validation")
                    continue
                obj = self._load_from_dict(input_dict, target_class=tc)
                for fmt in self.output_formats:
                    output_file = f"{base}.{fmt}"
                    if fmt == "yaml":
                        yaml_dumper.dump(obj, to_file=output_file)
                    elif fmt == "json":
                        json_dumper.dump(obj, to_file=output_file)
                    elif fmt == "ttl":
                        rdflib_dumper.dump(obj, to_file=output_file, schemaview=sv, prefix_map=self.prefix_map)
                    else:
                        raise NotImplementedError(f"Cannot output in format: {fmt}")
                    summary.outputs.append(f"{stem}.{fmt}")

    def _load_from_dict(self, dict_obj: Any, target_class: Union[str, ElementName] = None) -> Any:
        """
        Load an object from a dict, using the target class to determine the type of object to create.

        TODO: move logic into core

        :param dict_obj:
        :param target_class:
        :return:
        """
        if not self.use_type_designators:
            return dict_obj
        sv = self.schemaview
        if target_class is None:
            target_class_names = [c.name for c in sv.all_classes().values() if c.tree_root]
            if len(target_class_names) != 1:
                raise ValueError(f"Cannot determine single target class, found: {target_class_names}")
            target_class = target_class_names[0]
        if isinstance(dict_obj, dict):
            if target_class not in sv.all_classes():
                raise ValueError(f"No such class as {target_class}")
            td_slot = sv.get_type_designator_slot(target_class) if target_class else None
            if td_slot:
                if td_slot.name in dict_obj:
                    target_class = dict_obj[td_slot.name]
            elif "@type" in dict_obj:
                target_class = dict_obj["@type"]
                del dict_obj["@type"]
            if ":" in target_class:
                target_classes = [c for c in sv.all_classes() if sv.get_uri(c) == target_class]
                if len(target_classes) != 1:
                    raise ValueError(f"Cannot find unique class for URI {target_class}; got: {target_classes}")
                target_class = target_classes[0]
            new_dict_obj = {}

            for k, v in dict_obj.items():
                if v is not None:
                    islot = sv.induced_slot(k, target_class)
                    # if slot is a dictionary, repeat key in dictionary value object
                    if islot.multivalued and islot.inlined and not islot.inlined_as_list:
                        (range_id_slot, range_simple_dict_value_slot, _) = get_range_associated_slots(
                            self.schemaview, islot.range
                        )
                        v_as_list = []
                        for ik, iv in v.items():
                            # simple dictionaries can be simply created
                            if range_simple_dict_value_slot is not None:
                                value = {range_id_slot.name: ik, range_simple_dict_value_slot.name: iv}
                            # other dictionaries => simply add the identifier to the dictionary
                            else:
                                value = iv
                                value[range_id_slot.name] = ik
                            v_as_list.append(value)
                        v2 = self._load_from_dict(v_as_list, target_class=islot.range)
                    else:
                        v2 = self._load_from_dict(v, target_class=islot.range)
                    new_dict_obj[k] = v2
            py_target_class = getattr(self.python_module, camelcase(target_class))
            return py_target_class(**new_dict_obj)
        elif isinstance(dict_obj, list):
            return [self._load_from_dict(x, target_class) for x in dict_obj]
        else:
            return dict_obj


@click.command(name="examples")
@click.option("--schema", "-s", required=True, help="Path to linkml schema yaml file")
@click.option("--prefixes", "-P", help="Path to prefixes")
@click.option("--input-directory", "-e", help="folder containing positive examples that MUST pass validation")
@click.option(
    "--counter-example-input-directory",
    "-N",
    help="folder containing counter examples that MUST fail validation",
)
@click.option(
    "--output-directory",
    "-d",
    required=True,
    help="folder containing positive examples that MUST pass validation",
)
@click.option(
    "--input-formats",
    "-f",
    multiple=True,
    default=["yaml"],
    show_default=True,
    help="Target formats to be converted to (yaml, json)",
)
@click.option(
    "--output-formats",
    "-t",
    multiple=True,
    default=["yaml", "json", "ttl"],
    show_default=True,
    help="Target formats to be converted to (yaml, json, ttl)",
)
@click.option(
    "--output",
    "-o",
    default=sys.stdout,
    type=click.File("w", encoding="utf-8"),
    help="Output file for markdown summary",
)
@click.option(
    "--use-type-designators/--no-use-type-designators",
    default=True,
    show_default=True,
    help="If true use type_designators to deepen ranges",
)
@click.version_option(__version__, "-V", "--version")
def cli(schema, prefixes, output: TextIO, **kwargs):
    """Process a folder of examples and a folder of counter examples.

    Each example in the folder

    For context, see: https://github.com/linkml/linkml/issues/501
    """
    schemaview = SchemaView(schema)
    prefix_map = yaml.safe_load(open(prefixes)) if prefixes else None
    runner = ExampleRunner(schemaview=schemaview, prefix_map=prefix_map, **kwargs)
    runner.process_examples()
    output.write(str(runner.summary))


if __name__ == "__main__":
    cli(sys.argv[1:])



from linkml.generators.pydanticgen.pydanticgen import (
    DEFAULT_IMPORTS,
    MetadataMode,
    PydanticGenerator,
    cli,
)
from linkml.generators.pydanticgen.template import (
    ConditionalImport,
    Import,
    Imports,
    PydanticAttribute,
    PydanticBaseModel,
    PydanticClass,
    PydanticEnum,
    PydanticModule,
    PydanticTemplateModel,
    PydanticValidator,
)

__all__ = [
    "cli",
    "ConditionalImport",
    "DEFAULT_IMPORTS",
    "Import",
    "Imports",
    "MetadataMode",
    "PydanticAttribute",
    "PydanticBaseModel",
    "PydanticClass",
    "PydanticEnum",
    "PydanticGenerator",
    "PydanticModule",
    "PydanticValidator",
    "PydanticTemplateModel",
]



"""Generate JSONld from a LinkML schema."""

import os
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Optional

import click
from jsonasobj2 import as_json, items, loads
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    ElementName,
    SchemaDefinition,
    SlotDefinition,
    SlotDefinitionName,
    SubsetDefinition,
    SubsetDefinitionName,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.yamlutils import YAMLRoot

from linkml import METAMODEL_CONTEXT_URI
from linkml._version import __version__
from linkml.generators.jsonldcontextgen import ContextGenerator
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class JSONLDGenerator(Generator):
    """
    Generates JSON-LD from a Schema

    Status: incompletely implemented

    Note: this is distinct from
    :class:`~linkml.generators.jsonldcontextgen.ContextGenerator`, which generates a JSON-LD context
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.2"
    valid_formats = [
        "jsonld",
        "json",
    ]  # jsonld includes @type and @context.  json is pure JSON
    uses_schemaloader = True
    requires_metamodel = True
    file_extension = "jsonld"

    # ObjectVars
    original_schema: SchemaDefinition = None
    """See https://github.com/linkml/linkml/issues/871"""

    context: str = None
    """Path to a JSONLD context file"""

    def __post_init__(self) -> None:
        self.original_schema = deepcopy(self.schema)
        super().__post_init__()

    def _add_type(self, node: YAMLRoot) -> dict:
        if self.format == "jsonld":
            typ = node.__class__.__name__
            node = node.__dict__
            node["@type"] = typ
        return node

    def _visit(self, node: Any) -> Optional[Any]:
        if isinstance(node, (YAMLRoot, dict)):
            if isinstance(node, YAMLRoot):
                node = self._add_type(node)
            for k, v in list(items(node)):
                if v:
                    new_v = self._visit(v)
                    if new_v is not None:
                        node[k] = new_v
        elif isinstance(node, list):
            for i in range(0, len(node)):
                new_v = self._visit(node[i])
                if new_v is not None:
                    node[i] = new_v
        elif isinstance(node, set):
            for v in list(node):
                new_v = self._visit(v)
                if new_v is not None:
                    node.remove(v)
                    node.add(new_v)
        elif isinstance(node, ClassDefinitionName):
            return ClassDefinitionName(camelcase(node))
        elif isinstance(node, SlotDefinitionName):
            return SlotDefinitionName(underscore(node))
        elif isinstance(node, TypeDefinitionName):
            return TypeDefinitionName(underscore(node))
        elif isinstance(node, SubsetDefinitionName):
            return SubsetDefinitionName(underscore(node))
        elif isinstance(node, ElementName):
            return (
                ClassDefinitionName(camelcase(node))
                if node in self.schema.classes
                else (
                    SlotDefinitionName(underscore(node))
                    if node in self.schema.slots
                    else (
                        SubsetDefinitionName(camelcase(node))
                        if node in self.schema.subsets
                        else TypeDefinitionName(underscore(node))
                        if node in self.schema.types
                        else None
                    )
                )
            )
        return None

    def adjust_slot(self, slot: SlotDefinition) -> None:
        if slot.range in self.schema.classes:
            slot.range = ClassDefinitionName(camelcase(slot.range))
        elif slot.range in self.schema.slots:
            slot.range = SlotDefinitionName(underscore(slot.range))
        elif slot.range in self.schema.types:
            slot.range = TypeDefinitionName(underscore(slot.range))
        slot.slot_uri = self.namespaces.uri_for(slot.slot_uri)
        for f in [
            "mappings",
            "exact_mappings",
            "broad_mappings",
            "close_mappings",
            "narrow_mappings",
            "related_mappings",
        ]:
            setattr(slot, f, [self.namespaces.uri_for(v) for v in getattr(slot, f)])

    def visit_class(self, cls: ClassDefinition) -> bool:
        self._visit(cls)
        cls.class_uri = self.namespaces.uri_for(cls.class_uri)
        # Slot usage is a construction artifact
        # TODO: Figure out why this is here.  It isn't good form to alter a schema that may be used by other things
        cls.slot_usage = {}
        return False

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        self._visit(slot)
        self.adjust_slot(slot)

    def visit_type(self, typ: TypeDefinition) -> None:
        self._visit(typ)
        typ.uri = self.namespaces.uri_for(typ.uri)

    def visit_subset(self, ss: SubsetDefinition) -> None:
        self._visit(ss)

    def end_schema(self, context: str = None, **_) -> str:
        self._add_type(self.schema)
        base_prefix = self.default_prefix()

        # TODO: fix this, see https://github.com/linkml/linkml/issues/871
        # JSON LD adjusts context reference using '@base'.  If context is supplied and not a URI, generate an
        # absolute URI for it
        if context is None and self.format == "jsonld":
            # TODO: Once we get pyld running w/ relative contexts, we need to figure out how to generate and add
            #       the relative (?) context reference below
            # model_context = self.schema.source_file.replace('.yaml', '.prefixes.context.jsonld')
            # context = [METAMODEL_CONTEXT_URI, f'file://./{model_context}']
            # TODO: The _visit function above alters the schema in situ
            add_prefixes = ContextGenerator(self.original_schema, model=False, emit_metadata=False).serialize()
            add_prefixes_json = loads(add_prefixes)
            context = [METAMODEL_CONTEXT_URI, add_prefixes_json["@context"]]
        elif isinstance(context, str):  # Some of the older code doesn't do multiple contexts
            context = [context]
        elif isinstance(context, tuple):
            context = list(context)
        for imp in list(self.loaded.values())[1:]:
            context.append(imp[0] + ".context.jsonld")

        # Absolute file paths have to have a prefix
        for ci in range(0, len(context)):
            if isinstance(context[ci], str) and context[ci].startswith(
                "/"
            ):  # TODO: how do we deal with absolute DOS paths?
                context[ci] = "file://" + context[ci]

        if self.format == "jsonld":
            self.schema["@context"] = context[0] if len(context) == 1 and not base_prefix else context
            if base_prefix:
                self.schema["@context"].append({"@base": base_prefix})
        # json_obj["@id"] = self.schema.id
        out = str(as_json(self.schema, indent="  ")) + "\n"
        self.schema = self.original_schema
        return out


@shared_arguments(JSONLDGenerator)
@click.command(name="jsonld")
@click.option(
    "--context",
    multiple=True,
    help=f"JSONLD context file (default: {METAMODEL_CONTEXT_URI} and <model>.prefixes.context.jsonld)",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **kwargs):
    """Generate JSONLD file from LinkML schema.

    Status: incomplete
    """
    print(JSONLDGenerator(yamlfile, **kwargs).serialize(**kwargs))


if __name__ == "__main__":
    cli()



from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any


class Loader(ABC):
    """Abstract base class for instance data loaders.

    Subclasses must implement the iter_instances method.

    :param source: Path or URL to load instances from
    """

    def __init__(self, source) -> None:
        self.source = source

    @abstractmethod
    def iter_instances(self) -> Iterator[Any]:
        """Lazily load data instances from the source

        :return: Iterator over data instances
        :rtype: Iterator[Any]
        """
        pass



import sys
from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    ClassVar,
    Optional,
    TypeVar,
    Union,
)

from linkml_runtime.linkml_model import Element
from linkml_runtime.linkml_model.meta import ArrayExpression, DimensionExpression
from pydantic import VERSION as PYDANTIC_VERSION

from linkml.utils.deprecation import deprecation_warning

if int(PYDANTIC_VERSION[0]) < 2:
    # Support for having pydantic 1 installed in the same environment will be dropped in 1.9.0
    deprecation_warning("pydantic-v1")

if sys.version_info.minor < 12:
    from typing_extensions import TypeAliasType
else:
    from typing import TypeAliasType


from linkml.generators.pydanticgen.build import RangeResult
from linkml.generators.pydanticgen.template import ConditionalImport, Import, Imports, ObjectImport
from linkml.utils.exceptions import ValidationError


class ArrayRepresentation(Enum):
    LIST = "list"
    NUMPYDANTIC = "numpydantic"  # numpydantic must be installed to use this


_BOUNDED_ARRAY_FIELDS = ("exact_number_dimensions", "minimum_number_dimensions", "maximum_number_dimensions")

_T = TypeVar("_T")
AnyShapeArray = TypeAliasType("AnyShapeArray", list[Union[_T, "AnyShapeArray[_T]"]], type_params=(_T,))

_AnyShapeArrayImports = (
    Imports()
    + Import(
        module="typing",
        objects=[
            ObjectImport(name="TypeVar"),
            ObjectImport(name="Union"),
        ],
    )
    + ConditionalImport(
        condition="sys.version_info.minor >= 12",
        module="typing",
        objects=[ObjectImport(name="TypeAliasType")],
        alternative=Import(module="typing_extensions", objects=[ObjectImport(name="TypeAliasType")]),
    )
)

# annotated types are special and inspect.getsource() can't stringify them
_AnyShapeArrayInjects = [
    '_T = TypeVar("_T")',
    """AnyShapeArray = TypeAliasType(
    "AnyShapeArray", list[Union[_T, "AnyShapeArray[_T]"]], type_params=(_T,)
)""",
]

_ConListImports = Imports() + Import(module="pydantic", objects=[ObjectImport(name="conlist")])


class ArrayValidator:
    """
    Validate the specification of a LinkML Array

    .. todo::

        It looks like :mod:`linkml.validator` is for validating instances against schema, rather
        than validating the schema itself, so am not subclassing/writing as a plugin.
        Unsure if there is a more general means of validating schema, but for now this is
        an independent class
    """

    @classmethod
    def validate(cls, array: ArrayExpression):
        """
        Validate an array expression.

        Raises:
            :class:`.ValidationError` if invalid
        """
        cls.array_exact_dimensions(array)
        cls.array_consistent_n_dimensions(array)
        cls.array_explicitly_unbounded(array)
        cls.array_dimensions_ordinal(array)

        if array.dimensions:
            for dimension in array.dimensions:
                cls.validate_dimension(dimension)

    @classmethod
    def validate_dimension(cls, dimension: DimensionExpression):
        """
        Validate a single array dimension

        Raises:
            :class:`.ValidationError` if invalid
        """
        cls.dimension_exact_cardinality(dimension)
        cls.dimension_ordinal(dimension)

    @staticmethod
    def array_exact_dimensions(array: ArrayExpression):
        """Arrays can have exact_number_dimensions OR min/max_number_dimensions, but not both"""
        if array.exact_number_dimensions is not None and (
            array.minimum_number_dimensions is not None or array.maximum_number_dimensions is not None
        ):
            raise ValidationError(
                f"Can only specify EITHER exact_number_dimensions OR minimum/maximum dimensions, got: {array}"
            )

    @staticmethod
    def array_consistent_n_dimensions(array: ArrayExpression):
        """
        Complex arrays with both exact/min/max_number_dimensions and parameterized dimensions
        need to have the exact/min/max_number_dimensions greater than the number of parameterized dimensions!
        """
        if not array.dimensions:
            return

        for field_name in _BOUNDED_ARRAY_FIELDS:
            field = getattr(array, field_name, None)
            if field and field < len(array.dimensions):
                raise ValidationError(
                    "if exact/minimum/maximum_number_dimensions is provided, "
                    "it must be greater than the parameterized dimensions. "
                    f"got\n- {field_name}: {field}\n- dimensions: {array.dimensions}"
                )

    @staticmethod
    def array_dimensions_ordinal(array: ArrayExpression):
        """
        minimum_number_dimensions needs to be less than maximum_number_dimensions when both are set
        """
        if array.minimum_number_dimensions is not None and array.maximum_number_dimensions:
            if array.minimum_number_dimensions > array.maximum_number_dimensions:
                raise ValidationError(
                    "minimum_number_dimensions must be lesser than maximum_number_dimensions when both are set. "
                    f"got minimum: {array.minimum_number_dimensions}, maximum: {array.maximum_number_dimensions}"
                )

    @staticmethod
    def array_explicitly_unbounded(array: ArrayExpression):
        """
        Complex arrays with a minimum_number_dimensions and parameterized dimensions
        need to either use exact_number_dimensions to specify extra anonymous dimensions
        or set maximum_number_dimensions to ``False`` to specify unbounded extra anonymous
        dimensions to avoid ambiguity.
        """
        if array.minimum_number_dimensions is not None and array.maximum_number_dimensions is None and array.dimensions:
            raise ValidationError(
                "Cannot specify a minimum_number_dimensions while maximum is None while using labeled dimensions - "
                "either use exact_number_dimensions > len(dimensions) for extra parameterized dimensions or set "
                "maximum_number_dimensions explicitly to False for unbounded dimensions"
            )

    @staticmethod
    def dimension_exact_cardinality(dimension: DimensionExpression):
        """Dimensions can only have exact_cardinality OR min/max_cardinality, but not both"""
        if dimension.exact_cardinality is not None and (
            dimension.minimum_cardinality is not None or dimension.maximum_cardinality is not None
        ):
            raise ValidationError(
                f"Can only specify EITHER exact_cardinality OR minimum/maximum cardinality, got: {dimension}"
            )

    @staticmethod
    def dimension_ordinal(dimension: DimensionExpression):
        """minimum_cardinality must be less than maximum_cardinality when both are set"""
        if dimension.minimum_cardinality is not None and dimension.maximum_cardinality is not None:
            if dimension.minimum_cardinality > dimension.maximum_cardinality:
                raise ValidationError(
                    "minimum_cardinality must be lesser than maximum_cardinality when both are set. "
                    f"got minimum: {dimension.minimum_cardinality}, maximum: {dimension.maximum_cardinality}"
                )


class ArrayRangeGenerator(ABC):
    """
    Metaclass for generating a given format of array range.

    See :ref:`array-forms` for more details on array range forms.

    These classes do only enough validation of the array specification to decide
    which kind of representation to generate. Proper value validation should
    happen elsewhere (ie. in the metamodel and generated :class:`.ArrayExpression` class.)

    Each of the array representation generation methods should be able to handle
    the supported pydantic versions (currently still 1 and 2).

    Notes:

        When checking for array specification, recall that there is a semantic difference between
        ``None`` and ``False`` , particularly for :attr:`.ArrayExpression.max_number_dimensions` -
        check for absence of specification with ``is None`` rather than checking for truthiness/falsiness
        (unless that's what you intend to do ofc ;)

    Attributes:
        array (:class:`.ArrayExpression` ): Array to create a range for
        dtype (Union[str, :class:`.Element` ): dtype of the entire array as a string

    """

    REPR: ClassVar[ArrayRepresentation]

    def __init__(self, array: Optional[ArrayExpression], dtype: Union[str, Element]):
        self.array = array
        self.dtype = dtype

    def make(self) -> RangeResult:
        """
        Create the string form of the array representation, validating first
        """
        self.validate()

        if not self.array.dimensions and not self.has_bounded_dimensions:
            return self._any_shape(self.array)
        elif not self.array.dimensions and self.has_bounded_dimensions:
            return self._bounded_dimensions(self.array)
        elif self.array.dimensions and not self.has_bounded_dimensions:
            return self._parameterized_dimensions(self.array)
        else:
            return self._complex_dimensions(self.array)

    def validate(self):
        """
        Ensure that the given ArrayExpression is valid using :class:`.ArrayValidator`

        .. todo::

            Integrate with more general schema validation that happens when a schema is loaded,
            rather than when an array is generated

        Raises:
            :class:`.ValidationError` if the schema is invalid
        """
        ArrayValidator.validate(self.array)

    @property
    def has_bounded_dimensions(self) -> bool:
        """Whether the :class:`.ArrayExpression` has some shape specification aside from ``dimensions``"""
        return any([getattr(self.array, arr_field, None) is not None for arr_field in _BOUNDED_ARRAY_FIELDS])

    @classmethod
    def get_generator(cls, repr: ArrayRepresentation) -> type["ArrayRangeGenerator"]:
        """Get the generator class for a given array representation"""
        for subclass in cls.__subclasses__():
            if repr in (subclass.REPR, subclass.REPR.value):
                return subclass
        raise ValueError(f"Generator for array representation {repr} not found!")

    @abstractmethod
    def _any_shape(self, array: Optional[ArrayRepresentation] = None) -> RangeResult:
        """Any shaped array!"""
        pass

    @abstractmethod
    def _bounded_dimensions(self, array: ArrayExpression) -> RangeResult:
        """Array shape specified numerically, without axis parameterization"""
        pass

    @abstractmethod
    def _parameterized_dimensions(self, array: ArrayExpression) -> RangeResult:
        """Array shape specified with ``dimensions`` without additional parameterized dimensions"""
        pass

    @abstractmethod
    def _complex_dimensions(self, array: ArrayExpression) -> RangeResult:
        """Array shape with both ``parameterized`` and ``bounded`` dimensions"""
        pass


class ListOfListsArray(ArrayRangeGenerator):
    """
    Represent arrays as lists of lists!
    """

    REPR = ArrayRepresentation.LIST

    @staticmethod
    def _list_of_lists(dimensions: int, dtype: str) -> str:
        return ("list[" * dimensions) + dtype + ("]" * dimensions)

    @staticmethod
    def _parameterized_dimension(dimension: DimensionExpression, dtype: str) -> RangeResult:
        # TODO: Preserve label representation in some readable way! doing the MVP now of using conlist
        if dimension.exact_cardinality:
            dmin = dimension.exact_cardinality
            dmax = dimension.exact_cardinality
        elif dimension.minimum_cardinality or dimension.maximum_cardinality:
            dmin = dimension.minimum_cardinality
            dmax = dimension.maximum_cardinality
        else:
            # TODO: handle labels for labeled but unshaped arrays
            return RangeResult(range="list[" + dtype + "]")

        items = []
        if dmin is not None:
            items.append(f"min_length={dmin}")
        if dmax is not None:
            items.append(f"max_length={dmax}")

        items.append(f"item_type={dtype}")
        items = ", ".join(items)
        range = f"conlist({items})"

        return RangeResult(range=range, imports=_ConListImports)

    def _any_shape(self, array: Optional[ArrayExpression] = None, with_inner_union: bool = False) -> RangeResult:
        """
        An AnyShaped array (using :class:`.AnyShapeArray` )

        Args:
            array (:class:`.ArrayExpression`): The array expression (not used)
            with_inner_union (bool): If ``True`` , the innermost type is a ``Union`` of the ``AnyShapeArray`` class
                and ``dtype`` (default: ``False`` )

        """
        if self.dtype in ("Any", "AnyType"):
            range = "AnyShapeArray"
        else:
            range = f"AnyShapeArray[{self.dtype}]"

        if with_inner_union:
            range = f"Union[{range}, {self.dtype}]"
        return RangeResult(range=range, injected_classes=_AnyShapeArrayInjects, imports=_AnyShapeArrayImports)

    def _bounded_dimensions(self, array: ArrayExpression) -> RangeResult:
        """
        A nested series of ``list[]`` ranges with :attr:`.dtype` at the center.

        When an array expression allows for a range of dimensions, each set of ``List`` s is joined by a ``Union`` .
        """
        if array.exact_number_dimensions or (
            array.minimum_number_dimensions
            and array.maximum_number_dimensions
            and array.minimum_number_dimensions == array.maximum_number_dimensions
        ):
            exact_dims = array.exact_number_dimensions or array.minimum_number_dimensions
            return RangeResult(range=self._list_of_lists(exact_dims, self.dtype))
        elif not array.maximum_number_dimensions and (
            array.minimum_number_dimensions is None or array.minimum_number_dimensions == 1
        ):
            return self._any_shape()
        elif array.maximum_number_dimensions:
            # e.g., if min = 2, max = 3, range = Union[list[list[dtype]], list[list[list[dtype]]]]
            min_dims = array.minimum_number_dimensions if array.minimum_number_dimensions is not None else 1
            ranges = [self._list_of_lists(i, self.dtype) for i in range(min_dims, array.maximum_number_dimensions + 1)]
            return RangeResult(range="Union[" + ", ".join(ranges) + "]")
        else:
            # min specified with no max
            # e.g., if min = 3, range = list[list[AnyShapeArray[dtype]]]
            return RangeResult(
                range=self._list_of_lists(array.minimum_number_dimensions - 1, self._any_shape().range),
                injected_classes=_AnyShapeArrayInjects,
                imports=_AnyShapeArrayImports,
            )

    def _parameterized_dimensions(self, array: ArrayExpression) -> RangeResult:
        """
        Constrained shapes using :func:`pydantic.conlist`

        TODO:
        - preservation of aliases
        - (what other metadata is allowable on labeled dimensions?)
        """
        # generate dimensions from inside out and then format
        # e.g., if dimensions = [{min_card: 3}, {min_card: 2}],
        # range = conlist(min_length=3, item_type=conlist(min_length=2, item_type=dtype))
        range = self.dtype
        for dimension in reversed(array.dimensions):
            range = self._parameterized_dimension(dimension, range).range

        return RangeResult(range=range, imports=_ConListImports)

    def _complex_dimensions(self, array: ArrayExpression) -> RangeResult:
        """
        Mixture of parameterized dimensions with a max or min (or both) shape for anonymous dimensions.

        A mixture of ``List`` , :class:`.conlist` , and :class:`.AnyShapeArray` .
        """
        res = None
        # first process any unlabeled dimensions which must be the innermost level of the range,
        # then wrap that with labeled dimensions
        if array.exact_number_dimensions or (
            array.minimum_number_dimensions
            and array.maximum_number_dimensions
            and array.minimum_number_dimensions == array.maximum_number_dimensions
        ):
            exact_dims = array.exact_number_dimensions or array.minimum_number_dimensions
            if exact_dims > len(array.dimensions):
                res = RangeResult(range=self._list_of_lists(exact_dims - len(array.dimensions), self.dtype))
            elif exact_dims == len(array.dimensions):
                # equivalent to labeled shape
                return self._parameterized_dimensions(array)
            # else is invalid, see: ArrayValidator.array_consistent_n_dimensions

        elif array.maximum_number_dimensions is not None and not array.maximum_number_dimensions:
            # unlimited n dimensions, so innermost is AnyShape with dtype
            res = self._any_shape(with_inner_union=True)

            if array.minimum_number_dimensions:
                # some minimum anonymous dimensions but unlimited max dimensions
                # e.g., if min = 3, len(dim) = 2, then res.range = list[Union[AnyShapeArray[dtype], dtype]]
                # res.range will be wrapped with the 2 labeled dimensions later
                res.range = self._list_of_lists(array.minimum_number_dimensions - len(array.dimensions), res.range)

        elif array.maximum_number_dimensions:
            initial_min = array.minimum_number_dimensions if array.minimum_number_dimensions is not None else 0
            dmin = max(len(array.dimensions), initial_min) - len(array.dimensions)
            dmax = array.maximum_number_dimensions - len(array.dimensions)

            res = self._bounded_dimensions(
                ArrayExpression(minimum_number_dimensions=dmin, maximum_number_dimensions=dmax)
            )

        if res is None:
            raise ValueError("Unsupported array specification! this is almost certainly a bug!")  # pragma: no cover

        # Wrap inner dimension with labeled dimension
        # e.g., if dimensions = [{min_card: 3}, {min_card: 2}]
        # and res.range = list[Union[AnyShapeArray[dtype], dtype]]
        # (min 3 dims, no max dims)
        # then the final range = conlist(
        #     min_length=3,
        #     item_type=conlist(
        #         min_length=2,
        #         item_type=list[Union[AnyShapeArray[dtype], dtype]]
        #     )
        # )
        for dim in reversed(array.dimensions):
            res = res.merge(self._parameterized_dimension(dim, dtype=res.range))

        return res


class NumpydanticArray(ArrayRangeGenerator):
    """
    Represent array range with :class:`numpydantic.NDArray` annotations,
    allowing an abstract array specification to be used with many different array
    libraries.
    """

    REPR = ArrayRepresentation.NUMPYDANTIC
    MIN_NUMPYDANTIC_VERSION = "1.6.1"
    """
    Minimum numpydantic version needed to be installed in the environment using
    the generated models
    """
    IMPORTS = Imports() + Import(
        module="numpydantic", objects=[ObjectImport(name="NDArray"), ObjectImport(name="Shape")]
    )
    INJECTS = [f'MIN_NUMPYDANTIC_VERSION = "{MIN_NUMPYDANTIC_VERSION}"']

    def make(self) -> RangeResult:
        result = super().make()
        result.imports = self.IMPORTS.model_copy()
        result.injected_classes = self.INJECTS.copy()
        return result

    @staticmethod
    def ndarray_annotation(shape: Optional[list[Union[int, str]]] = None, dtype: Optional[str] = None) -> str:
        """
        Make a stringified :class:`numpydantic.NDArray` annotation for a given shape
        and dtype.

        If either ``shape`` or ``dtype`` is ``None`` , use ``Any``
        """
        if shape is None:
            shape = "Any"
        else:
            shape_expression = ", ".join([str(i) for i in shape])
            shape = f'Shape["{shape_expression}"]'

        if dtype is None or dtype in ("Any", "AnyType"):
            dtype = "Any"

        if shape == "Any" and dtype == "Any":
            return "NDArray"
        else:
            return f"NDArray[{shape}, {dtype}]"

    @staticmethod
    def _dimension_shape(dimension: DimensionExpression) -> str:
        if dimension.exact_cardinality:
            shape = str(dimension.exact_cardinality)
        elif dimension.minimum_cardinality and not dimension.maximum_cardinality:
            shape = f"{dimension.minimum_cardinality}-*"
        elif dimension.maximum_cardinality and not dimension.minimum_cardinality:
            shape = f"*-{dimension.maximum_cardinality}"
        elif dimension.minimum_cardinality and dimension.maximum_cardinality:
            shape = f"{dimension.minimum_cardinality}-{dimension.maximum_cardinality}"
        else:
            shape = "*"

        return shape

    @classmethod
    def _parameterized_dimension(cls, dimension: DimensionExpression) -> str:
        shape = cls._dimension_shape(dimension)
        if dimension.alias is not None:
            return f"{shape} {dimension.alias}"
        else:
            return shape

    def _any_shape(self, array: Optional[ArrayRepresentation] = None) -> RangeResult:
        """
        Any shaped array, either an unparameterized :class:`numpydantic.NDArray`
        if dtype is :class:`typing.Any` , or like ``NDArray[Any, {self.dtype}]``
        otherwise.
        """
        if self.dtype in ("Any", "AnyType"):
            range = "NDArray"
        else:
            range = f"NDArray[Any, {self.dtype}]"

        return RangeResult(range=range)

    def _bounded_dimensions(self, array: ArrayExpression) -> RangeResult:
        """
        Number of dimensions specified without shape
        """
        if array.exact_number_dimensions or (
            array.minimum_number_dimensions
            and array.maximum_number_dimensions
            and array.minimum_number_dimensions == array.maximum_number_dimensions
        ):
            exact_dims = array.exact_number_dimensions or array.minimum_number_dimensions

            return RangeResult(range=self.ndarray_annotation(["*"] * exact_dims, self.dtype))
        elif not array.maximum_number_dimensions and (
            array.minimum_number_dimensions is None or array.minimum_number_dimensions == 1
        ):
            return self._any_shape()
        elif array.maximum_number_dimensions:
            # e.g., if min = 2, max = 3, range = Union[NDArray[Shape["*, *"], dtype], NDArray[Shape["*, *, *"], dtype]]
            min_dims = array.minimum_number_dimensions if array.minimum_number_dimensions is not None else 1
            ranges = [
                self.ndarray_annotation(["*"] * i, self.dtype)
                for i in range(min_dims, array.maximum_number_dimensions + 1)
            ]
            return RangeResult(range="Union[" + ", ".join(ranges) + "]")
        else:
            # min specified with no max
            # e.g., if min = 3, range = NDArray[Shape[*, *, *, ...], dtype]
            shape_inner = ["*"] * array.minimum_number_dimensions
            shape_inner.append("...")
            return RangeResult(range=self.ndarray_annotation(shape_inner, self.dtype))

    def _parameterized_dimensions(self, array: ArrayExpression) -> RangeResult:
        """
        Arrays with constrained shapes or labels
        """
        dims = [self._parameterized_dimension(d) for d in array.dimensions]
        range = self.ndarray_annotation(dims, self.dtype)
        return RangeResult(range=range)

    def _complex_dimensions(self, array: ArrayExpression) -> RangeResult:
        """
        Mixture of parameterized dimensions with a max or min (or both) shape for anonymous dimensions.
        """
        dims = [self._parameterized_dimension(d) for d in array.dimensions]
        res = None

        if array.exact_number_dimensions or (
            array.minimum_number_dimensions
            and array.maximum_number_dimensions
            and array.minimum_number_dimensions == array.maximum_number_dimensions
        ):
            exact_dims = array.exact_number_dimensions or array.minimum_number_dimensions
            if exact_dims > len(array.dimensions):
                dims.extend(["*"] * (exact_dims - len(dims)))
                res = self.ndarray_annotation(dims, self.dtype)
            elif exact_dims == len(array.dimensions):
                # equivalent to labeled shape
                return self._parameterized_dimensions(array)
            # else is invalid, see: ArrayValidator.array_consistent_n_dimensions(array)

        elif array.maximum_number_dimensions is not None and not array.maximum_number_dimensions:
            # unlimited n dimensions

            if array.minimum_number_dimensions:
                # some minimum anonymous dimensions but unlimited max dimensions
                dims.extend(["*"] * (array.minimum_number_dimensions - len(dims)))

            dims.append("...")
            res = self.ndarray_annotation(dims, self.dtype)

        elif array.maximum_number_dimensions:
            # some res of anonymous dimensions

            if array.minimum_number_dimensions:
                min_dim = array.minimum_number_dimensions
            else:
                min_dim = len(dims)

            dim_union = []
            for i in range(min_dim, array.maximum_number_dimensions + 1):
                this_dims = dims.copy()
                this_dims.extend(["*"] * (i - len(dims)))
                dim_union.append(self.ndarray_annotation(this_dims, self.dtype))
            dim_union = ", ".join(dim_union)
            res = f"Union[{dim_union}]"

        if res is None:
            raise ValueError(f"Unhandled range case! {array}")

        return RangeResult(range=res)



import logging
import os
from dataclasses import dataclass
from typing import Union

import click
from linkml_runtime.dumpers import json_dumper, yaml_dumper
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments
from linkml.utils.helpers import write_to_file

logger = logging.getLogger(__name__)

# type annotation for file name
FILE_TYPE = Union[str, bytes, os.PathLike]


@dataclass
class LinkmlGenerator(Generator):
    """This generator provides a direct conversion of a LinkML schema
    into json, optionally merging imports and unrolling induced slots
    into attributes
    """

    generatorname = os.path.basename(__file__)
    generatorversion = "1.0.0"
    valid_formats = ["json", "yaml"]
    uses_schemaloader = False
    requires_metamodel = False

    materialize_attributes: bool = False
    materialize_patterns: bool = False

    def __post_init__(self):
        # TODO: consider moving up a level
        super().__post_init__()
        self.schemaview = SchemaView(self.schema, merge_imports=self.mergeimports)

    def materialize_classes(self) -> None:
        """Materialize class slots from schema as attributes, in place"""
        all_classes = self.schemaview.all_classes()

        for c_name, c_def in all_classes.items():
            attrs = self.schemaview.class_induced_slots(c_name)
            for attr in attrs:
                c_def.attributes[attr.name] = attr

    def serialize(self, output: str = None, **kwargs) -> str:
        if self.materialize_attributes:
            self.materialize_classes()
        if self.materialize_patterns:
            self.schemaview.materialize_patterns()

        if self.format == "json":
            json_str = json_dumper.dumps(self.schemaview.schema)

            if output:
                write_to_file(output, json_str)
                logger.info(f"Materialized file written to: {output}")
                return output

            return json_str
        elif self.format == "yaml":
            yaml_str = yaml_dumper.dumps(self.schemaview.schema)

            if output:
                write_to_file(output, yaml_str)
                logger.info(f"Materialized file written to: {output}")
                return output

            return yaml_str
        else:
            raise ValueError(
                f"{self.format} is an invalid format. Use one of the following formats: {self.valid_formats}"
            )


@shared_arguments(LinkmlGenerator)
@click.option(
    "--materialize/--no-materialize",
    default=True,
    show_default=True,
    help="Materialize both, induced slots as attributes and structured patterns as patterns",
)
@click.option(
    "--materialize-attributes/--no-materialize-attributes",
    default=True,
    show_default=True,
    help="Materialize induced slots as attributes",
)
@click.option(
    "--materialize-patterns/--no-materialize-patterns",
    default=True,
    show_default=True,
    help="Materialize structured patterns as patterns",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(),
    help="Name of JSON or YAML file to be created",
)
@click.version_option(__version__, "-V", "--version")
@click.command(name="linkml")
def cli(
    yamlfile,
    materialize: bool,
    materialize_attributes: bool,
    materialize_patterns: bool,
    output: FILE_TYPE = None,
    **kwargs,
):
    # You can use the `--materialize` / `--no-materialize` for control
    # over both attribute and pattern materialization.

    # If the user did not explicitly specify materialize_attributes,
    # fall back to the umbrella materialize flag.
    if materialize_attributes is None:
        materialize_attributes = materialize

    # If the user did not explicitly specify materialize_patterns,
    # fall back to the umbrella materialize flag.
    if materialize_patterns is None:
        materialize_patterns = materialize

    gen = LinkmlGenerator(
        yamlfile,
        materialize_attributes=materialize_attributes,
        materialize_patterns=materialize_patterns,
        **kwargs,
    )
    print(gen.serialize(output=output))


if __name__ == "__main__":
    cli()



from collections.abc import Iterator
from typing import Any

from linkml.validator.loaders.loader import Loader


class PassthroughLoader(Loader):
    """A loader that passes through from an existing Iterator

    :param source: An Iterator
    """

    def __init__(self, source: Iterator) -> None:
        super().__init__(source)

    def iter_instances(self) -> Iterator[Any]:
        """Pass through instances from an Iterator

        :return: Iterator over data instances
        :rtype: Iterator[Any]
        """
        yield from self.source



import logging

import click

LOG_LEVEL_STRINGS = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
DEFAULT_LOG_LEVEL: str = "WARNING"
DEFAULT_LOG_LEVEL_INT: int = logging.WARNING


def log_level_option(fn):
    def callback(ctx, param, value):
        log_level_string = value.upper()
        log_level_int = getattr(logging, log_level_string, None)
        if not isinstance(log_level_int, int):
            raise ValueError(f"Invalid log level: {log_level_string}")
        logging.basicConfig(level=log_level_int)

    return click.option(
        "--log_level",
        type=click.Choice(LOG_LEVEL_STRINGS),
        help="Logging level",
        default=DEFAULT_LOG_LEVEL,
        show_default=True,
        callback=callback,
    )(fn)



"""
Base classes for jinja template handling classes.

See :mod:`.linkml.generators.pydanticgen.template` for example implementation
"""

from collections.abc import Generator
from copy import copy
from typing import Any, ClassVar, Optional, Union

from jinja2 import Environment
from pydantic import BaseModel, Field, field_validator


class TemplateModel(BaseModel):
    """
    Metaclass to render model results with jinja templates.

    Each subclass needs to declare a :class:`typing.ClassVar` for a
    jinja template within the `templates` directory.

    Templates are written expecting each of the other TemplateModels
    to already be rendered to strings - ie. rather than the ``class.py.jinja``
    template receiving a full :class:`.PydanticAttribute` object or dictionary,
    it receives it having already been rendered to a string. See the :meth:`.render` method.
    """

    template: ClassVar[str]
    _environment: ClassVar[Environment]

    meta_exclude: ClassVar[list[str]] = None

    def render(self, environment: Optional[Environment] = None, **kwargs) -> str:
        """
        Recursively render a template model to a string.

        For each field in the model, recurse through, rendering each :class:`.TemplateModel`
        using the template set in :attr:`.TemplateModel.template` , but preserving the structure
        of lists and dictionaries. Regular :class:`.BaseModel` s are rendered to dictionaries.
        Any other value is passed through unchanged.

        Args:
            environment (:class:`jinja2.Environment`): Template environment - see :meth:`.environment`
        """
        if environment is None:
            environment = type(self).environment()

        fields = {**self.model_fields, **self.model_computed_fields}

        data = {k: _render(getattr(self, k, None), environment) for k in fields}
        template = environment.get_template(self.template)
        rendered = template.render(**data)
        return rendered

    @classmethod
    def environment(cls) -> Environment:
        """
        Default environment for Template models.
        uses a :class:`jinja2.PackageLoader` for the templates directory within this module
        with the ``trim_blocks`` and ``lstrip_blocks`` parameters set to ``True`` so that the
        default templates could be written in a more readable way.
        """
        return copy(cls._environment)

    @classmethod
    def exclude_from_meta(cls: "TemplateModel") -> list[str]:
        """
        Attributes in the source definition to exclude from linkml_meta
        """
        ret = [*cls.model_fields.keys()]
        if cls.meta_exclude is not None:
            ret = ret + cls.meta_exclude
        return ret


class ObjectImport(BaseModel):
    """
    An object to be imported from within a module.

    See :class:`.Import` for examples
    """

    name: str
    alias: Optional[str] = None


class Import(TemplateModel):
    """
    A module or module and classes to be imported.
    """

    module: str
    alias: Optional[str] = None
    objects: Optional[list[ObjectImport]] = None
    is_schema: bool = False
    """
    Whether or not this ``Import`` is importing another schema imported by the main schema --
    ie. that it is not expected to be provided by the environment, but imported locally from within the package.
    Used primarily in split schema generation, see :func:`.pydanticgen.generate_split` for example usage.
    """

    def merge(self, other: "Import") -> list["Import"]:
        """
        Merge one import with another, see :meth:`.Imports` for an example.

        * If module don't match, return both
        * If one or the other are a :class:`.ConditionalImport`, return both
        * If modules match, neither contain objects, but the other has an alias, return the other
        * If modules match, one contains objects but the other doesn't, return both
        * If modules match, both contain objects, merge the object lists, preferring objects with aliases
        """
        # return both if we are orthogonal
        if self.module != other.module:
            return [self, other]

        # handle conditionals
        if isinstance(self, ConditionalImport) and isinstance(other, ConditionalImport):
            # If our condition is the same, return the newer version
            if self.condition == other.condition:
                return [other]
        if isinstance(self, ConditionalImport) or isinstance(other, ConditionalImport):
            # we don't have a good way of combining conditionals, just return both
            return [self, other]

        # handle module vs. object imports
        elif other.objects is None and self.objects is None:
            # both are modules, return the other only if it updates the alias
            if other.alias:
                return [other]
            else:
                return [self]
        elif other.objects is not None and self.objects is not None:
            # both are object imports, merge and return
            alias = self.alias if other.alias is None else other.alias
            # FIXME: super awkward implementation
            # keep ours if it has an alias and the other doesn't,
            # otherwise take the other's version
            self_objs = {obj.name: obj for obj in self.objects}
            other_objs = {
                obj.name: obj for obj in other.objects if obj.name not in self_objs or self_objs[obj.name].alias is None
            }
            self_objs.update(other_objs)

            return [
                type(self)(
                    module=self.module,
                    alias=alias,
                    objects=list(self_objs.values()),
                    is_schema=self.is_schema or other.is_schema,
                    **{
                        k: getattr(other, k)
                        for k in other.model_fields
                        if k not in ("module", "alias", "objects", "is_schema")
                    },
                )
            ]
        else:
            # one is a module, the other imports objects, keep both
            return [self, other]

    def sort(self) -> None:
        """
        Sort imported objects

        * First by whether the first letter is capitalized or not,
        * Then alphabetically (by object name rather than alias)
        """
        if self.objects:
            self.objects = sorted(self.objects, key=lambda obj: (obj.name[0].islower(), obj.name))


class ConditionalImport(Import):
    """
    Import that depends on some condition in the environment, common when
    using backported features or straddling dependency versions.

    Make sure that everything that is needed to evaluate the condition is imported
    before this is added to the injected imports!
    """

    condition: str
    alternative: Import

    def sort(self) -> None:
        """
        :meth:`.Import.sort` called for self and :attr:`.alternative`
        """
        super().sort()
        self.alternative.sort()


class Imports(TemplateModel):
    """
    Container class for imports that can handle merging!

    See :class:`.Import` and :class:`.ConditionalImport` for examples of declaring individual imports

    Useful for generation, because each build stage will potentially generate
    overlapping imports. This ensures that we can keep a collection of imports
    without having many duplicates.

    Defines methods for adding, iterating, and indexing from within the :attr:`Imports.imports` list.
    """

    imports: list[Union[Import, ConditionalImport]] = Field(default_factory=list)
    render_sorted: bool = True
    """When rendering, render in sorted groups"""

    @classmethod
    def _merge(
        cls, imports: list[Union[Import, ConditionalImport]], other: Union[Import, "Imports", list[Import]]
    ) -> list[Union[Import, ConditionalImport]]:
        """
        Add a new import to an existing imports list, handling deduplication and flattening.

        Mutates and returns ``imports``

        Generally will prefer the imports in ``other`` , updating those in ``imports``.
        If ``other`` ...
        - doesn't match any ``module`` in ``imports``, add it!
        - matches a single ``module`` in imports, :meth:`.Import.merge` the object imports
        - matches multiple ``module``s in imports, then there must have been another
            :class:`.ConditionalImport` already present, so we :meth:`.Import.merge` the existing
            :class:`.Import` if it is one, and if it's a :class:`.ConditionalImport` just YOLO
            and append it since there isn't a principled way to merge them from strings.
        - is :class:`.Imports`  or a list of :class:`.Import` s, call this recursively for each
          item.

        Since imports can be merged in an undefined order depending on the generator configuration,
        default behavior for imports with matching ``module`` is to remove them and append to the
        end of the imports list (rather than keeping it in the position of the existing
        :class:`.Import` ). :class:`.ConditionalImports` make it possible to have namespace
        conflicts, so in imperative import style we assume the most recently added :class:`.Import`
        is the one that should prevail.
        """
        #
        if isinstance(other, cls) or (isinstance(other, list) and all([isinstance(i, Import) for i in other])):
            for i in other:
                imports = cls._merge(imports, i)
            return imports

        existing = [i for i in imports if i.module == other.module]

        # if we have nothing importing from this module yet, add it!
        if len(existing) == 0:
            imports.append(other)
        elif len(existing) == 1:
            imports.remove(existing[0])
            imports.extend(existing[0].merge(other))
        else:
            # we have both a conditional and at least one nonconditional already.
            # If this is another conditional, we just add it, otherwise, we merge it
            # with the single nonconditional
            if isinstance(other, ConditionalImport):
                imports.append(other)
            else:
                for e in existing:
                    if isinstance(e, Import):
                        imports.remove(e)
                        merged = e.merge(other)
                        imports.extend(merged)
                        break

        # SPECIAL CASE - __future__ annotations must happen at the top of a file
        # sort here outside of sort method because our imports are invalid without it,
        # where calling ``sort`` should be optional.
        imports = sorted(imports, key=lambda i: i.module == "__future__", reverse=True)
        return imports

    def __add__(self, other: Union[Import, "Imports", list[Import]]) -> "Imports":
        imports = self.imports.copy()
        imports = self._merge(imports, other)
        return type(self).model_construct(
            imports=imports, **{k: getattr(self, k, None) for k in self.model_fields if k != "imports"}
        )

    def __len__(self) -> int:
        return len(self.imports)

    def __iter__(self) -> Generator[Import, None, None]:
        yield from self.imports

    def __getitem__(self, item: Union[int, str]) -> Import:
        if isinstance(item, int):
            return self.imports[item]
        elif isinstance(item, str):
            # the name of the module
            an_import = [i for i in self.imports if i.module == item]
            if len(an_import) == 0:
                raise KeyError(f"No import with module {item} was found.\nWe have: {self.imports}")
            return an_import[0]
        else:
            raise TypeError(f"Can only index with an int or a string as the name of the module,\nGot: {type(item)}")

    def __contains__(self, item: Union[Import, "Imports", list[Import]]) -> bool:
        """
        Check if all the objects are imported from the given module(s)

        If the import is a bare module import (ie its :attr:`~.Import.objects` is ``None`` )
        then we must also have a bare module import in this Imports (because even if
        we import from the module, unless we import it specifically its name won't be
        available in the namespace.

        :attr:`.Import.alias` must always match for the same reason.
        """
        if isinstance(item, Imports):
            return all([i in self for i in item.imports])
        elif isinstance(item, list):
            return all([i in self for i in item])
        elif isinstance(item, Import):
            try:
                an_import = self[item.module]
            except KeyError:
                return False
            if item.objects is None:
                return an_import == item
            else:
                return all([obj in an_import.objects for obj in item.objects])
        else:
            raise TypeError(f"Imports only contains single Import objects or other Imports\nGot: {type(item)}")

    @field_validator("imports", mode="after")
    @classmethod
    def imports_are_merged(
        cls, imports: list[Union[Import, ConditionalImport]]
    ) -> list[Union[Import, ConditionalImport]]:
        """
        When creating from a list of imports, construct model as if we have done so by iteratively
        constructing with __add__ calls
        """
        merged_imports = []
        for i in imports:
            merged_imports = cls._merge(merged_imports, i)
        return merged_imports

    def sort(self) -> None:
        """
        Sort imports recursively, mimicking isort:

        * First by :attr:`.Import.group` according to :attr:`.Imports.group_order`
        * Then by whether the :class:`.Import` has any objects
          (``import module`` comes before ``from module import name``)
        * Then alphabetically by module name
        """

        def _sort_key(i: Import) -> tuple[int, str]:
            return (int(i.objects is not None), i.module)

        imports = sorted(self.imports, key=_sort_key)
        for i in imports:
            i.sort()
        self.imports = imports

    def render(self, environment: Optional[Environment] = None, black: bool = False) -> str:
        if self.render_sorted:
            self.sort()
        rendered = super().render(environment=environment, black=black)
        return rendered


def _render(
    item: Union[TemplateModel, Any, list[Union[Any, TemplateModel]], dict[str, Union[Any, TemplateModel]]],
    environment: Environment,
) -> Union[str, list[str], dict[str, str]]:
    if isinstance(item, TemplateModel):
        return item.render(environment)
    elif isinstance(item, list):
        return [_render(i, environment) for i in item]
    elif isinstance(item, dict):
        return {k: _render(v, environment) for k, v in item.items()}
    elif isinstance(item, BaseModel):
        fields = {**item.model_fields, **getattr(item, "model_computed_fields", {})}
        return {k: _render(getattr(item, k, None), environment) for k in fields.keys()}
    else:
        return item



"""
Utilities for formatting pydanticgen outputs

Kept separate in case we decide we don't want it/we want it to be moved to a place
that's better for other generators to use.
"""

from typing import Optional

try:
    from black import Mode, TargetVersion, format_str
except ImportError:
    import warnings

    warnings.warn("Black is an optional dependency, install it with the extra 'black' like `pip install linkml[black]`")


def _default_mode() -> "Mode":
    return Mode(target_versions={TargetVersion.PY311}, line_length=120)


def format_black(code: str, mode: Optional["Mode"] = None) -> str:
    if mode is None:
        mode = _default_mode()

    formatted = format_str(code, mode=mode)
    return formatted
